'''
 Generated by EZX Code Generator  
Date: 04/18/22 
'''


from iserver.EzxMsg import EzxMsg, EzxMsgMeta
from iserver.msgs import *
from iserver.enums import api




class PairsSpreadOrder(EzxMsg, metaclass=EzxMsgMeta):


	def __init__(self, **kwargs):
		self.msgType : int = None
		self.myID : str = None
		self.routerOrderID : int = None
		self.userNum : int = None
		self.userName : str = None
		self.side : int = None
		self.price : float = None
		self.orderQty : float = None
		self.icebergQty : int = None
		self.fairValue : float = None
		self.priceOffset : float = None
		self.xSymbol : str = None
		self.ySymbol : str = None
		self.xMdSymbol : str = None
		self.yMdSymbol : str = None
		self.xText : str = None
		self.yText : str = None
		self.destination : str = None
		self.account : str = None
		self.locateReqd : str = None
		self.hangAllowance : int = None
		self.moveAllowance : float = None
		self.slipAllowance : float = None
		self.xSlipAllowance : float = None
		self.ySlipAllowance : float = None
		self.yPriceWeight : float = None
		self.yHedgeWeight : float = None
		self.initialScan : float = None
		self.hedgeScan : float = None
		self.replaceThrottleTime : float = None
		self.maxDistanceFromMarket : float = None
		self.orderDelayMs : int = None
		self.idSourceX : str = None
		self.securityIdX : str = None
		self.idSourceY : str = None
		self.securityIdY : str = None
		super(PairsSpreadOrder, self).__init__(**kwargs)
		self.msg_type = api.IserverMsgType.APPLICATION.value
		self.msg_subtype = api.IserverMsgSubType.PAIRS_SPREAD_ORDER.value




'''
 Generated by EZX Code Generator  
Date: 04/18/22 
'''

from enum import Enum, unique

@unique
class IserverMsgType(Enum):
	DISCONNECTED = 6
	SYNC = 1
	COMMAND = 3
	SESSION = 5
	UNDEFINED = 0
	SYNC_COMPLETE = 4
	APPLICATION = 2



@unique
class IserverMsgSubType(Enum):
	CLIENT_SETTINGS_UPDATE = 146
	READ_GROUP_ACCOUNTS = 136
	PARENT_ORDER_REQUEST = 115
	ORDERINFO_REQUEST = 105
	BROADCAST_RESPONSE = 125
	REJECT = 111
	DKTRADE = 124
	PARENT_ORDER_RESPONSE = 116
	UPDATE_GROUP_SETTINGS = 142
	FORGOTPASSWORD_RESPONSE = 135
	EXECUTION = 103
	CHANGEPASSWORD = 132
	STATUS = 120
	REPLACE_REJECT = 118
	DELETE_GROUP = 141
	ADMIN_RESPONSE = 10
	UPDATE_POSITION = 144
	DESTINATION_STATUS = 127
	ORDERREQUESTLIST = 38
	QUOTE_ACCEPT = 155
	GET_ORDERS_RESPONSE = 96
	MOVE_ACCOUNT = 140
	FIXEDINCOMEMARKETDATAINCREMENTAL = 148
	ALLOCINFO_RESPONSE = 92
	EXECUTION_RESPONSE = 108
	TEST_CHILD_CLASS = 152
	ORDERINFO_RESPONSE = 106
	CHANGEPASSWORD_RESPONSE = 133
	EXECUTION_REQUEST = 107
	UPDATE_GROUP_NAME = 143
	MARKETDATASNAPSHOT = 36
	ADMIN_COMMAND = 126
	QUOTE = 128
	FORGOTPASSWORD = 134
	POSITION_RESPONSE = 145
	ALLOCATION_RESPONSE = 94
	PAIRS_SPREAD_ORDER_RESPONSE = 150
	CANCEL_ALL = 11
	CLIENT_ORDER_RESPONSE = 99
	HEARTBEAT = 122
	GET_ORDERS_REQUEST = 95
	LOGONRESPONSE = 110
	CANCEL_REJECT = 119
	ORDERS = 113
	ORDERRESPONSE = 112
	BROADCAST = 104
	MARKETDATAINCREMENTAL = 37
	CREDIT_USED_UPDATE = 147
	CANCEL = 117
	COMPANY_SETTINGS_UPDATE = 131
	EXECUTIONS = 114
	UNSPECIFIED = 0
	TEST_PARENT_CLASS = 151
	CREATE_GROUP_ACCOUNT = 138
	ORDER = 102
	OMSSETTINGS = 97
	LOGON = 100
	QUOTE_RESPONSE = 154
	PAIRS_SPREAD_ORDER = 149
	CLIENT_ORDER_REQUEST = 98
	QUOTE_REQUEST = 129
	ALLOCINFO_REQUEST = 91
	MARKETDATA_HEALTH = 153
	DELETE_CLIENT_ORDER = 130
	ALLOCATION_REQUEST = 93
	DELETE_ACCOUNT = 139
	REPLACE = 121
	LOGOUT = 101
	IOI = 39






'''
Created on 11 Mar 2022

@author: shalomshachne
'''
import time

from iserver.EzxMsg import EzxMsg
from iserver.enums.msgenums import State, Side
from _datetime import datetime


__counter = 0

closed_states = { 
    State.REJT.value,
    State.CAND.value,
    State.FILLED.value,
    State.EXPIRE.value,
    State.OVRF.value, 
    State.DONE.value
    }


def next_id() -> str:
    '''
    Get next unique ID. Uses a combination of millis time and a counter. Not thread safe.
    '''
    global __counter
    if __counter >= 999:
        __counter = 0
    __counter = __counter + 1
    millis = round(time.time() * 1000)
    return str(millis) + '-' + str(__counter)


def is_closed(order : EzxMsg):
    try:
        return order.state in closed_states
    except: 
        # if wrong kind of object gets here and returns false
        pass

def format_order(order: EzxMsg):
    # Buy IBM 100x165.23 (ROID=1234,ssOrderID=45667)
    side = Side(order.side).name    
    return f'{side} {order.symbol} {order.orderQty}x{order.price: .2f} (ROID={order.routerOrderID}) ({order.event}) (Filled={order.cumQty})'
      
def start_timing():
    watch = StopWatch()
    watch.start()
    return watch
      

class StopWatch(object):    
    
    def __init__(self):
        self.__start_time = None
        self.__stop_time = None
        self.__elapsed = None
        
    
    def start(self):
        self.__start_time = time.monotonic_ns()
        self.__stop_time = None
        self.__elapsed = None        

    def stop(self):
        self.__stop_time = time.monotonic_ns()
        self.__elapsed = self.__stop_time - self.__start_time  # difference in nanos
        
    def is_running(self):
        return self.__start_time and not self.__stop_time
        
    def elapsed_nanos(self) -> int:
        if self.__elapsed:
            return self.__elapsed
        
        if self.__start_time:
            return time.monotonic_ns() - self.__start_time
        
        return 0
         
    def elapsed_micros(self) -> float:
        return self.elapsed_nanos() / 1000
        
    def elapsed_millis(self) -> float:
        return self.elapsed_nanos() / 1000000
    
    def elapsed_seconds(self) -> float:
        return self.elapsed_nanos() / 1000000000
    
    def reset(self):
        self.__init__()
    


from _io import StringIO
import io

from iserver import msg_factory
from iserver.EzxMsg import EzxMsg
from iserver.msgs import msgtags

# API constants for delimiters in messages
TAG_VALUE_PAIRS_DELIMITER = '\001'
TAG_VALUE_SEPARATOR = '='
END_MESSAGE_DELIMITER = '\003'

# API doubles are sent as strings. Max decimals transmitted = 8
_DOUBLE_FORMAT = ".8g"          


def zero_pad(value, digits):
    return str(value).zfill(digits)



def to_string(value):
    if isinstance(value, float):
        return format(value, _DOUBLE_FORMAT)
    return str(value)


def msg_to_map(msg):
    d = dict()
    if len(msg) > 0 and msg[-1] == END_MESSAGE_DELIMITER:
        msg = msg[0:-1]
    for pair in msg.split(TAG_VALUE_PAIRS_DELIMITER):
        if TAG_VALUE_SEPARATOR in pair:
            d.update([pair.split(TAG_VALUE_SEPARATOR)])
    return d

'''
Decode received text to the appropriate API message (as determined by the msg_subtype value). If invalid msg_subtype,
returns None (and does not throw an error).
 
Created on Oct 5, 2021

@author: Sgoldberg
'''  


def decode_message(msg_subtype: int, msg: str):
    '''
    Decode a message received from the API
    @param msg_subtype: int which defines what type of message was received.  See IserverMsgSubType
    @param msg: a text message (decoded from bytes received from in the ApiClient socket). 
    '''
    m = msg_factory.get_message(msg_subtype)    
    if not m:
        return
    
    return populate_message(m, msg)
#     for tag_values in msg.split(TAG_VALUE_PAIRS_DELIMITER):
#         msg_field = None        
#         for item in tag_values.split(TAG_VALUE_SEPARATOR):
#             if not msg_field:
#                 msg_field = msgtags.get_msg_field(item, m)
#                 if not msg_field:
#                     break
#             else:
#                 msg_field.update_msg_field(m, item)
#             
#     return m    


def populate_message(m: EzxMsg, msg: str, start_index: int=None, end_index: int=None):
    '''
    Populate the specified EzxMsg object with values read from the received message string. For a message string within the message, 
    specify start_index and end_index parameters to parse the submessage only.
    
    @param m: a message object to populate
    @param msg: text received to decode into the message object
    @param start_index: optional index of where to start parsing the message string (defaults to beginning of the string)
    @param end_index: optional end index where to end parsing (defaults to end of the string)    
    '''
    tag_buffer = StringIO()
    value_buffer = StringIO()
    tag = None
    value = None
    
    msg_field = None     
    i = start_index or 0   
    stop = end_index or len(msg)
    while i < stop:
        c = msg[i]
        # increment here (TODO: explain why - has to do with return from parse_block)
        i = i + 1   
        
        # it's possible that a state variable would make this easier to understand. There are 2 states:
        #    1. reading the Tag name - signaled by start of the message or completing reading last value (hit the \x001 delimiter
        #    2. reading the value - signaled by reading an '=' sign at the end of the tag name.
        
        if c == TAG_VALUE_SEPARATOR and tag == None:  # if we already have the tag defined, then this '=' character is inside the value
            tag = tag_buffer.getvalue()      
            msg_field = msgtags.get_msg_field(tag, m)            
            
        elif c == TAG_VALUE_PAIRS_DELIMITER:
            value = value_buffer.getvalue()
            if 'B' == tag:  # indicates we have a sub-message (message block/group) contained within the message.
                i, *_ = parse_block(i, msg, m, value)  # read just the submessage section and continue at the next tag not in the submessage
                        
            elif msg_field: 
                msg_field.update_msg_field(m, value)
            
            else:  # last chance method, used to update custom messages
                m.update(tag, value)

            # reset buffers for next tag/value                
            tag_buffer.close()
            tag_buffer = StringIO()
            tag = None
            value_buffer.close()
            value_buffer = StringIO()
            msg_field = None
            
        elif msg_field == None and tag == None:
            tag_buffer.write(c)
            
        else:
            value_buffer.write(c)
                
    return m   

'''
Parse subsequence of the message into another message object and update the parent message object 
with the result.
@param start_index: start index for in message
@param msg: the text received from API
@param msg_object: the parent msg which contains the message block (sub message) 
@param block_id:   the string which identifies what kind of message the block represents
@return: return index in string for next character to read.   
'''


def parse_block(start_index: int, msg: str, msg_object: EzxMsg, block_id: str) -> int:
    end_text = f'E={block_id}\001'
    block_end = msg.find(end_text, start_index) 
    # TODO: handle bad message where block_end = -1
    block_msg = msg_factory.get_block_message(block_id, msg_object)    
    if block_msg:
        populate_message(block_msg, msg, start_index, block_end) 
        block_field = msg_object.get_block_field(block_id)
        if block_field:
            block_field.update_msg_field(msg_object, block_msg)
        # TODO: figure out how to update the base message with the block message
        # e.g. msg_factory.get_block_field(msg_object, block_id)
            
    return block_end + len(end_text), block_msg


def populate_message2(m, msg):
    for tag_values in msg.split(TAG_VALUE_PAIRS_DELIMITER):
        msg_field = None        
        for item in tag_values.split(TAG_VALUE_SEPARATOR):
            if not msg_field:
                msg_field = msgtags.get_msg_field(item, m)
                if not msg_field:
                    break
            else:
                msg_field.update_msg_field(m, item)
            
    return m    

        
def decode_block(block_tag: str, msg_fragment: str):
    m = msg_factory.get_block_message(block_tag)
    if not m:
        return
    

def parse_header(header: bytearray):
    # TODO: check length?
    header = header.decode('UTF-8')
    msg_length = int(header[0:6])
    msg_type = int(header[6:8])
    msg_subtype = int(header[8:11])
    return (msg_length, msg_type, msg_subtype)
    
    
def encode(msg: EzxMsg) -> str: 
    '''
    Encode the specified message a text string of name/value pairs (according to the API encoding) 
    '''
    msg_buffer = io.StringIO()    
    encode_buffer(msg, msg_buffer)
    return msg_buffer.getvalue()


def encode_buffer(msg: str, buffer: StringIO):
    '''
        Encode the specified message a text string of name/value pairs (according to the API encoding) into the specified StringIO buffer  
    '''
    for field_name, value in vars(msg).items(): 
        if value == None:
            continue
        isBlock = True
        f = msg.get_block_field_by_name(field_name)
        if not f:
            f = msgtags.get_api_field(field_name)
            isBlock = False
        if not f:
            continue
        
        encode_field(msg, buffer, f, value, isBlock)


def encode_tag_value(buffer, value, tag):
    buffer.write(tag)
    buffer.write(TAG_VALUE_SEPARATOR)
    buffer.write(to_string(value))
    buffer.write(TAG_VALUE_PAIRS_DELIMITER)


def encode_field(msg, buffer, field, value, isBlock):
        if (type(value) == list):
            for item in value: encode_field(msg, buffer, field, item, isBlock)
            return
        
        # dictionary is a field in a custom message (TagValueMsg) it is not a standard API object type (yet)
        if (type(value) == dict):
            for tag, item_value in value.items():
                encode_tag_value(buffer, item_value, tag)
            return
            
        if isBlock:
            # the value is a message, write the Start block Tag/Value
            encode_tag_value(buffer, field.api_tag, 'B')
            # encode the full message
            encode_buffer(value, buffer)
            # write the End block Tag/Value            
            encode_tag_value(buffer, field.api_tag, 'E')
            return
                        
        tag = field.api_tag
        encode_tag_value(buffer, value, tag)        

    
def header(msg_length: int, msg_type: int, msg_subtype: int, seqno: int, buffer: io.StringIO):
    buffer.write(zero_pad(msg_length + 1, 6))  # add ETX character
    buffer.write(zero_pad(msg_type, 2))
    buffer.write(zero_pad(msg_subtype, 3))
    buffer.write(zero_pad(seqno, 3))
    buffer.write(TAG_VALUE_PAIRS_DELIMITER)




def to_api_msg(msg: EzxMsg, seqno: int=0) -> str:
    '''
    Generate an API message from the specified object. This will start with 15 byte 
    API Header and end with API END_MESSAGE_DELIMITED (Ascii 3)
    '''        
    api_buffer = io.StringIO() 
    msg_buffer = io.StringIO()           
    encode_buffer(msg, msg_buffer)
    # print(f'msg_buffer={msg_buffer.getvalue()}')
    header(msg_buffer.tell(), msg.msg_type, msg.msg_subtype, seqno, api_buffer)
    api_buffer.write(msg_buffer.getvalue())
    api_buffer.write(END_MESSAGE_DELIMITER)
    return api_buffer.getvalue()


def to_api_bytes(msg: EzxMsg, seqno: int=0) -> bytearray:
    return to_api_msg(msg, seqno).encode(encoding='UTF-8')
    





'''
Created on 11 Feb 2022

@author: shalomshachne
'''
import unittest
from iserver.msgs.ExecutionStatus import ExecutionStatus
from iserver.msgs.ClientOrderResponse import ClientOrderResponse


class Test(unittest.TestCase):


    def setUp(self):
        self.response = ClientOrderResponse()


    def tearDown(self):
        pass


    def testKwArgsConstructor(self):
        e = ExecutionStatus()
        self.response.execStatus = e
        self.response.symbol = "XYZ"
        
        response2 = ClientOrderResponse(execStatus = e, symbol = self.response.symbol)
        self.assertEquals(self.response, response2)
        self.assertEqual(e, response2.execStatus, "set the ExecStatus property")
        
        



'''
 Generated by EZX Code Generator  
Date: 04/18/22 
'''


from iserver.EzxMsg import EzxMsg, EzxMsgMeta
from iserver.msgs import *
from iserver.enums import api




class ExecutionRequest(EzxMsg, metaclass=EzxMsgMeta):


	def __init__(self, **kwargs):
		self.execTransType : int = None
		self.execType : int = None
		self.orderQty : float = None
		self.ordType : str = None
		self.exeTimeStamp : str = None
		self.price : float = None
		self.lastPrice : float = None
		self.lastShares : float = None
		self.routerOrderID : int = None
		self.ssOrderID : int = None
		self.ssExeID : int = None
		self.trader : str = None
		self.text : str = None
		self.userInfo : str = None
		super(ExecutionRequest, self).__init__(**kwargs)
		self.msg_type = api.IserverMsgType.APPLICATION.value
		self.msg_subtype = api.IserverMsgSubType.EXECUTION_REQUEST.value




'''
 Generated by EZX Code Generator  
Date: 04/18/22 
'''


from iserver.EzxMsg import EzxMsg, EzxMsgMeta
from iserver.msgs import *
from iserver.enums import api
from iserver.msgs.PairsSpreadOrder import PairsSpreadOrder




class PairsSpreadOrderResponse(PairsSpreadOrder, metaclass=EzxMsgMeta):


	def __init__(self, **kwargs):
		self.filledQty : int = None
		self.hungQty : int = None
		self.xFilledQty : int = None
		self.yFilledQty : int = None
		self.xAvgPrice : float = None
		self.yAvgPrice : float = None
		self.pairState : str = None
		self.state : str = None
		self.event : str = None
		self.cumQty : float = None
		self.leavesQty : float = None
		self.timeStamp : str = None
		self.text : str = None
		self.ownerID : str = None
		super(PairsSpreadOrderResponse, self).__init__(**kwargs)
		self.msg_type = api.IserverMsgType.APPLICATION.value
		self.msg_subtype = api.IserverMsgSubType.PAIRS_SPREAD_ORDER_RESPONSE.value




'''
Created on 16 Mar 2022
Subclasses of API messages tuned to specific use cases
@author: shalomshachne
'''
from iserver.msgs.OrderRequest import OrderRequest
from iserver.enums.msgenums import MsgType, OrdType


class NewOrder(OrderRequest):
    
    def __init__(self, symbol : str, side : int, orderQty : int, price : float = None, destination : str = None, myID : str = None, orderType : int = None, **kwargs):
        super().__init__(**kwargs)
        self.msgType = MsgType.NEW.value    
        self.symbol = symbol
        self.side = side
        self.orderQty = orderQty
        self.price = price
        self.ordType = orderType
        self.destination = destination
        self.myID = myID
        
        if not self.ordType:
            if price and price > 0:
                self.ordType = OrdType.LIMIT.value
            else:
                self.ordType = OrdType.MARKET.value
                self.price = None
        
        


class CancelOrder(OrderRequest):
    '''
    Convenience message to cancel an order. The only field which needs to be set is routerOrderID.  
    '''
    def __init__(self, routerOrderID: int):
        '''
        The constructor sets self.msgType = MsgType.CANC.value which is required by API to indicate that this is a cancel.
        @param routerOrderID: the order id of order to cancel 
        '''
        super().__init__()
        self.msgType = MsgType.CANC.value
        self.routerOrderID = routerOrderID

        
class ReplaceOrder(OrderRequest):
    '''
    Convenience message to replace an order. For a replace, the only required fields are the routerOrderID and *only* fields whose values need to be
    replaced. Fields whose values are not changing don't need to be set.
    '''
    
    def __init__(self, routerOrderID: int, newPrice: float=None, newQty: int=None, **kwargs):
        '''
        The constructor sets self.msgType = MsgType.REPL.value which is required by API to indicate that this is a replace.
        Replaces typically are for price only, although sometimes also for orderQty. Both arguments are optional. To set other
        order fields to replace, use the kwargs parameter. 
        
        @param routerOrderID: the order id of order to replace
        @param newPrice: new price desired (default=None)
        @param newQty: new orderQty desired (default=None)
        @param @**kwargs: to set other fields as needed 
        '''        
        super().__init__(**kwargs)
        self.msgType = MsgType.REPL.value
        self.routerOrderID = routerOrderID
        self.price = newPrice
        self.orderQty = newQty



'''
 Generated by EZX Code Generator  
Date: 04/18/22 
'''


from iserver.EzxMsg import EzxMsg, EzxMsgMeta
from iserver.msgs import *
from iserver.enums import api
from iserver.msgs.OrderStatus import OrderStatus

from .ParentOrderStatus import ParentOrderStatus
from .ExecutionStatus import ExecutionStatus
from .OrderRequestStatus import OrderRequestStatus
from .StrategyInfo import StrategyInfo



class OrderResponse(OrderStatus, metaclass=EzxMsgMeta):

	__block_fields__ = {
		'PB' : Field('parents', 'PB', noop_decode_func, update_list_item), 
		'EB' : Field('executions', 'EB', noop_decode_func, update_list_item), 
		'ORSB' : Field('orderRequests', 'ORSB', noop_decode_func, update_list_item), 
		'STGBLK' : Field('strategyInfo', 'STGBLK', noop_decode_func, ), 
	}

	def __init__(self, **kwargs):
		self.parents : list[ParentOrderStatus] = None
		self.executions : list[ExecutionStatus] = None
		self.orderRequests : list[OrderRequestStatus] = None
		self.strategyInfo : StrategyInfo = None
		super(OrderResponse, self).__init__(**kwargs)
		self.msg_type = api.IserverMsgType.APPLICATION.value
		self.msg_subtype = api.IserverMsgSubType.ORDERRESPONSE.value




'''
 Generated by EZX Code Generator  
Date: 04/18/22 
'''


from iserver.EzxMsg import EzxMsg, EzxMsgMeta
from iserver.msgs import *
from iserver.enums import api

from .AllocInfo import AllocInfo
from .AllocationRequestStatus import AllocationRequestStatus



class AllocationStatus(EzxMsg, metaclass=EzxMsgMeta):

	__block_fields__ = {
		'ALLOCINFO' : Field('allocInfoList', 'ALLOCINFO', noop_decode_func, update_list_item), 
		'ALLOCREQSB' : Field('allocationRequests', 'ALLOCREQSB', noop_decode_func, update_list_item), 
	}

	def __init__(self, **kwargs):
		self.myID : str = None
		self.routerAllocID : int = None
		self.version : int = None
		self.seqNo : int = None
		self.allocID : str = None
		self.allocStatus : str = None
		self.allocRejCode : int = None
		self.state : str = None
		self.text : str = None
		self.tradeDate : str = None
		self.clientNameID : int = None
		self.symbol : str = None
		self.symbolsfx : str = None
		self.side : int = None
		self.currency : str = None
		self.shares : int = None
		self.avgPx : float = None
		self.timeStamp : str = None
		self.allocInfoList : list[AllocInfo] = None
		self.allocationRequests : list[AllocationRequestStatus] = None
		super(AllocationStatus, self).__init__(**kwargs)




'''
 Generated by EZX Code Generator  
Date: 04/18/22 
'''


from iserver.EzxMsg import EzxMsg, EzxMsgMeta
from iserver.msgs import *
from iserver.enums import api




class ExecutionStatus(EzxMsg, metaclass=EzxMsgMeta):


	def __init__(self, **kwargs):
		self.event : str = None
		self.execTransType : int = None
		self.execType : int = None
		self.orderQty : float = None
		self.ordType : str = None
		self.exeTimeStamp : str = None
		self.exeVersion : int = None
		self.price : float = None
		self.lastPrice : float = None
		self.lastShares : float = None
		self.options : str = None
		self.senderSubID : str = None
		self.ssExeID : int = None
		self.tagValueBlock : str = None
		self.targetSubID : str = None
		self.text : str = None
		self.transactTime : str = None
		self.tradeLiquidityIndicator : str = None
		self.ecnAccessFee : float = None
		self.cfiCode : str = None
		self.refID : str = None
		self.ratioQty : int = None
		self.positionEffect : str = None
		self.lastMkt : str = None
		self.iServerGenerated : int = None
		self.clientOrderID : str = None
		self.origClientOrderID : str = None
		self.brokerOrderID : str = None
		super(ExecutionStatus, self).__init__(**kwargs)




'''
 Generated by EZX Code Generator  
Date: 04/18/22 
'''


from iserver.EzxMsg import EzxMsg, EzxMsgMeta
from iserver.msgs import *
from iserver.enums import api

from .TagValueMsg import TagValueMsg



class Properties(EzxMsg, metaclass=EzxMsgMeta):

	__block_fields__ = {
		'STRATSETTVS' : Field('propertyMap', 'STRATSETTVS', noop_decode_func, ), 
	}

	def __init__(self, **kwargs):
		self.companyName : str = None
		self.propertyMap : TagValueMsg = None
		super(Properties, self).__init__(**kwargs)




'''
 Generated by EZX Code Generator  
Date: 04/18/22 
'''


from iserver.EzxMsg import EzxMsg, EzxMsgMeta
from iserver.msgs import *
from iserver.enums import api

from .YieldData import YieldData
from .PartiesStatus import PartiesStatus
from .BondExecutionStatus import BondExecutionStatus



class QuoteResponse(EzxMsg, metaclass=EzxMsgMeta):

	__block_fields__ = {
		'YDB' : Field('yieldData', 'YDB', noop_decode_func, ), 
		'PRTS' : Field('parties', 'PRTS', noop_decode_func, update_list_item), 
		'BEB' : Field('bondExecutionStatus', 'BEB', noop_decode_func, ), 
	}

	def __init__(self, **kwargs):
		self.quoteID : str = None
		self.quoteEntryID : str = None
		self.quoteRespID : str = None
		self.seqNo : int = None
		self.account : str = None
		self.clientOrderID : str = None
		self.state : str = None
		self.price : float = None
		self.side : int = None
		self.text : str = None
		self.settlementType : int = None
		self.cumQty : float = None
		self.avgPrice : float = None
		self.ordType : str = None
		self.orderQty : float = None
		self.minQty : float = None
		self.maxFloor : float = None
		self.quoteCondition : str = None
		self.numCompetitors : int = None
		self.coverPrice : float = None
		self.quoteRank : str = None
		self.yieldData : YieldData = None
		self.parties : list[PartiesStatus] = None
		self.bondExecutionStatus : BondExecutionStatus = None
		super(QuoteResponse, self).__init__(**kwargs)
		self.msg_type = api.IserverMsgType.APPLICATION.value
		self.msg_subtype = api.IserverMsgSubType.QUOTE_RESPONSE.value




'''
Created on 18 Feb 2022

@author: SGoldberg
'''
from iserver.EzxMsg import EzxMsg
from iserver.msgs import Field, noop_decode_func


class TagValueMsg(EzxMsg):
    '''
    classdocs
    '''

    def __init__(self):
        '''
        Constructor
        '''
        self.tag_values : dict[str, str] = {}
    
    def update(self, tag:str, value:object):
        self.tag_values[tag] = value
        


'''
 Generated by EZX Code Generator  
Date: 04/18/22 
'''


from iserver.EzxMsg import EzxMsg, EzxMsgMeta
from iserver.msgs import *
from iserver.enums import api



# constants
STATE_DOWN = 'ENGINE_DOWN'
STATE_UP = 'ENGINE_UP'


class EngineStatus(EzxMsg, metaclass=EzxMsgMeta):


	def __init__(self, **kwargs):
		self.fixEngineName : str = None
		self.fixEngineState : str = None
		super(EngineStatus, self).__init__(**kwargs)
		self.msg_type = api.IserverMsgType.APPLICATION.value
		self.msg_subtype = api.IserverMsgSubType.STATUS.value




'''
Created on Sep 29, 2021

@author: Sgoldberg
'''
import io
from pickle import TRUE
import unittest

from iserver import ezx_msg
from iserver.msgs import LogonRequest
from iserver.msgs import LogonResponse
from tests.data import MsgWithBlockFields, GrandChild


class Parent(object):
    def __init__(self):
        self.parent_name = None
        

class Child(Parent):
    def __init__(self):
        super(Child, self).__init__()
        self.child_name = None
        
        




class Test(unittest.TestCase):
    
    def testParentChild(self):
        c = Child()
        self.assertTrue(hasattr(c, 'child_name'), 'child name')
        self.assertTrue(hasattr(c, 'parent_name'), 'parent name')
        
        
        

    def testListAttributes(self):
        l = LogonRequest()
        print(f'vars={vars(l)}')
        for name, value in vars(l).items():
            print(f"{name}={value}")


        
    def testDoubleToString(self):
        value = 123.45
        print(str(value))
        
        value = value + .01
        print(str(value))
   
    def test_to_str(self):     
        value = 123.45
        print(ezx_msg.to_string(value))
        
        value = value + .01
        print(ezx_msg.to_string(value))   
        
        value = 543210
        print(ezx_msg.to_string(value))
    
    def test_get_string_char(self):
        s = "Hello World"
        self.assertEqual('d', s[-1], "got last char")
        

    def test_filter(self):
        l = [1, 2, 3, 4, 5];
        f = lambda x : x % 2 == 0
        result = list(filter(f, l))
        print(f'result={result}')
        self.assertEqual(2, len(result))
        
    def test_attr_functions(self):
        msg = LogonResponse()
        name = 'returnCode'
        self.assertTrue(hasattr(msg, name))
        
        name = 'destinations'
        self.assertTrue(hasattr(msg, name))
        
        name = 'returnCode'
        value = getattr(msg, name)
        self.assertIsNone(value, f"get attr {name}")
        
        name = 'destinations'
        value = getattr(msg, name)
        self.assertIsNone(value, f"get attr {name}")
        
        
    # API types
    # list
    # double
    # GroupAccount
    # int
    # LocalTimeStamp
    # LogonType
    # String
    # UTCTimeStamp        
    def test_convert_api_field_value(self):
        s = "aString"
        t = str(s)
        self.assertEqual(s, t)
        
    
    def test_meta_class1(self):
        msg = MsgWithBlockFields()
        self.assertTrue(hasattr(msg, '__block_fields__'))
        self.assertTrue(hasattr(msg, '__block_fields_by_name__'))
        
    def test_meta_class_with_multiple_inheritance(self):
        msg = GrandChild()
        self.assertTrue(hasattr(msg, '__block_fields__'))
        self.assertTrue(hasattr(msg, '__block_fields_by_name__'))
        
    
    def test_lambda_with_closure(self):
        class Simple:
            def __init__(self):
                self.is_enabled = False  

        closure_obj = Simple()
        predicate = lambda : closure_obj.is_enabled
        
        self.assertFalse(predicate(), 'before change to object')
        closure_obj.is_enabled = True
        self.assertTrue(predicate(), 'lambda evaluated state of closure object')
        
    def test_memory_view_bytearray(self):
        size = 10
        b = bytearray(size)
        self.assertEqual(size, len(b))
        b[0] = 65
        b[1] = 66
        b[2] = 67
        s = b.decode('UTF-8').strip('\x00')
        self.assertEqual('ABC', s)
        
        index = 3
        b2 = memoryview(b)[index:]
        self.assertEqual(size - index, len(b2), 'created view of bytearray')
        
        char = 68
        for i in range(0,len(b2)):
            b2[i] = char
            char += 1
        
        print(f"b2={b2[0]}")
        self.assertTrue(b2.obj is b)
        expected = 'ABCDEFGHIJ'
        actual = b.decode('UTF-8')
        self.assertEqual(expected, actual, 'original buffer filled in as expected')
         
        
    def test_memory_view_copy(self):
        size = 10
        b = bytearray(size)
        self.assertEqual(size, len(b))
        b[0] = 65
        b[1] = 66
        b[2] = 67
        s = b.decode('UTF-8').strip('\x00')
        self.assertEqual('ABC', s)
        
        index = 3
        b2 = memoryview(b)[index:]
        view_size = len(b2)
        self.assertEqual(size - index, view_size, 'created view of bytearray')
        
        b3 = bytearray(view_size)
        b3 = b'DEFGHIJ'        
        b2[0:] = b3
  
        
        print(f"b2={b2[0]}")
        self.assertTrue(b2.obj is b)
        expected = 'ABCDEFGHIJ'
        actual = b.decode('UTF-8')
        self.assertEqual(expected, actual, 'original buffer filled in as expected')        
        
        
if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testListAttributes']
    unittest.main()



'''
 Generated by EZX Code Generator  
Date: 04/18/22 
'''


from iserver.EzxMsg import EzxMsg, EzxMsgMeta
from iserver.msgs import *
from iserver.enums import api

from .OrderRequestInfo import OrderRequestInfo



class OrderRequestList(EzxMsg, metaclass=EzxMsgMeta):

	__block_fields__ = {
		'ORB' : Field('orderList', 'ORB', noop_decode_func, update_list_item), 
	}

	def __init__(self, **kwargs):
		self.orderList : list[OrderRequestInfo] = None
		super(OrderRequestList, self).__init__(**kwargs)
		self.msg_type = api.IserverMsgType.APPLICATION.value
		self.msg_subtype = api.IserverMsgSubType.ORDERREQUESTLIST.value




'''
Created on Nov 9, 2021

@author: samgo_000
'''
import unittest
from iserver.enums.msgenums import State

from iserver.msgs import OrderResponse


class Test(unittest.TestCase):

    def testConstructorWithKeyWordArgs(self):
        sym = 'ABC'
        side = 2
        order_qty = 100
        price = 10
        or1 = OrderResponse()
        or1.symbol = sym
        or1.side = side
        or1.orderQty = order_qty
        or1.price = price
        
        or2 = OrderResponse(symbol=sym, side=side, orderQty=order_qty, price=price)
        self.assertEqual(or1, or2)
    
    def testEnumTypeInStateField(self):
        sym = 'ABC'
        side = 2
        order_qty = 100
        price = 10
        or1 = OrderResponse()
        or1.symbol = sym
        or1.side = side
        or1.orderQty = order_qty
        or1.price = price
        or1.state = 'ACKED'
        
        self.assertEqual(State.ACKED.value, or1.state, "state enum matches string value")
        
        

if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testConstructorWithKeyWordArgs']
    unittest.main()



'''
Created on 21 Mar 2022

@author: shalomshachne
'''
import logging
from time import sleep
import time
import unittest

from iserver.net import ClientMonitor, ApiClient, ClientState
from tests.data import test_data_factory
from tests import wait_for_condition


waitTime = .050
def wait_for_connection(client):
    start = time.time()
    maxwait = waitTime * 10 
    while not ClientState.CONNECTED == client.state() and time.time() - start <= maxwait:
        sleep(waitTime)

class ClientMonitorTest(unittest.TestCase):


    def setUp(self):
        self.client = MonitorTestClient()
        self.client.set_retry_seconds(waitTime)        
        self.monitor = ClientMonitor(self.client)
        

    def tearDown(self):
        self.monitor.stop()

    def test_initial_state(self):
        self.assertFalse(self.monitor.is_running(), 'not running')

    def testStart(self):
        logging.debug('testStart: starting monitor')        
        self.monitor.start()
        logging.debug('testStart: waiting for connection')        
        wait_for_connection(self.client)
        self.assertEqual(ClientState.CONNECTED, self.client.state(), 'called connect')
        self.assertTrue(self.monitor.is_running(), 'monitor is running')
        
    def test_stop(self):
        self.testStart()
        self.monitor.stop()
        wait_for_condition(lambda : not self.monitor.is_running())
        self.assertFalse(self.monitor.is_running(), 'correct running state')
        
        
    def testReconnects(self):
        logging.debug('testReconnects: started')
        self.testStart()
        self.assertEqual(1, self.client.connect_counter, 'valid initial state, 1 connection attempt')
        
        self.client._set_state(ClientState.DISCONNECTED)
        sleep(waitTime * 5)
        self.assertEqual(2, self.client.connect_counter, 'attempted reconnect')
        self.assertEqual(ClientState.CONNECTED, self.client.state(), 'called connect')

    def testNoReconnectWhenLoggedIn(self):
        self.testStart()
        self.client._set_state(ClientState.LOGGED_IN)
        sleep(waitTime * 2)
        self.assertEqual(1, self.client.connect_counter, 'no attempted reconnect')
    
    def testNoReconnectWhenStoped(self):
        self.testStart()
        self.client._set_state(ClientState.STOPPED)
        sleep(waitTime * 2)
        self.assertEqual(1, self.client.connect_counter, 'no attempted reconnect')
        
    
    def testSendsHeartbeatMessagesWhenLoggedIn(self):
        self.monitor.heartbeat_timer.start()
        
        hb_seconds = .030
        self.client.heartbeat_seconds = hb_seconds        
        self.client._set_state(ClientState.LOGGED_IN)
        self.monitor.start()
        sleep(hb_seconds * 2)
        self.assertTrue(self.client.heartbeat_count > 0, 'sent some heartbeats')
        
    def testHeartbeatSentAtIntervals(self):
        hb_seconds = .160
        self.client.heartbeat_seconds = hb_seconds
        
        self.testStart()
        self.assertFalse(self.client.heartbeat_count, 'no heartbeats sent')
        
        self.client._set_state(ClientState.LOGGED_IN)
        
        sleep(hb_seconds * 1.5)
        self.assertEqual(1, self.client.heartbeat_count, 'sent heartbeat after hb time')
        
        
        
        

class MonitorTestClient(ApiClient):
    def __init__(self):
        super().__init__(test_data_factory.create_connection_info())
        self.started = False
        self.connect_counter = 0
        self.heartbeat_count = 0
        
    def start(self):
        self.started = True
        
    def _connect(self):
        self.connect_counter = self.connect_counter + 1
        self._set_state(ClientState.CONNECTED)
        
    def set_retry_seconds(self, seconds : float):
        self.connection_info.connect_retry_seconds = seconds    
    
    @property
    def heartbeat_seconds(self):
        return self.connection_info.heartbeat_seconds   
    
    @heartbeat_seconds.setter
    def heartbeat_seconds(self, seconds: float):
        self.connection_info.heartbeat_seconds = seconds
        
    def _heartbeat(self):
        self.heartbeat_count += 1

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


'''
Created on Oct 29, 2021

@author: Sgoldberg
'''
import unittest

from iserver.EzxMsg import EzxMsg
from iserver.msgs import OrderResponse, Field, noop_decode_func, \
    update_list_item
from iserver.msgs.OrderRequest import OrderRequest


class MsgChild(EzxMsg):
    __block_fields__ = {
        'ORB' : Field('legList', 'ORB', noop_decode_func, update_list_item), 
        'ALLOCABLK' : Field('allocsList', 'ALLOCABLK', noop_decode_func, update_list_item), 
    }
    
class MsgGrandChild(MsgChild):
        __block_fields__ = {
        'STGBLK' : Field('legList', 'STGBLK', noop_decode_func, update_list_item)
    }

class MsgGreatGrandChild(MsgGrandChild):
        __block_fields__ = {
        'HOBO' : Field('legList', 'HOBO', noop_decode_func, update_list_item)
    }

class Test(unittest.TestCase):


    def testGetBlockFieldByNameFindsField(self):
        orderResponse = OrderResponse()
        f = orderResponse.get_block_field_by_name('executions')
        self.assertIsNotNone(f)
        self.assertEqual('executions', f.name)
        
    def testGetBlockFieldByNameFindsFieldWhenParent(self):
        order = OrderRequest();
        f = order.get_block_field_by_name('strategyInfo')
        self.assertIsNotNone(f)
        self.assertEqual('strategyInfo', f.name)

    
    def testGetBlockFieldChecksClassHeirarchy(self):
        # test that get_block_field navigates the message heirarchy.
        msg = MsgGreatGrandChild()
        toCheck = [ MsgGreatGrandChild, MsgGrandChild, MsgChild ]        
        for c in toCheck:
            obj = c()
            d = obj.__block_fields__
            for name in d:
                f = msg.get_block_field(name)
                self.assertIsNotNone(f, f'found field for {name}, originating from class={c}')
                
    def testGetBlockFieldOwner(self):
        msg = MsgGreatGrandChild()
        toCheck = [ MsgGreatGrandChild, MsgGrandChild, MsgChild ]        
        for c in toCheck:
            obj = c()
            d = obj.__block_fields__
            for name in d:
                found = msg.get_block_field_owner(name)
                self.assertEqual(c, found, f'got correct owner class for {name}')
                
    def testGetBlockFieldMethodsReturnNoneIfNotFound(self):
        msg = MsgGreatGrandChild()
        name = 'notexist'
        self.assertIsNone(msg.get_block_field_owner(name))
        self.assertIsNone(msg.get_block_field(name))
        self.assertIsNone(msg.get_block_field_by_name(name))
        

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testGetBlockFieldByName']
    unittest.main()


'''
 Generated by EZX Code Generator  
Date: 06/08/22 
'''


from iserver.EzxMsg import EzxMsg, EzxMsgMeta
from iserver.msgs import *
from iserver.enums import api

from .OrderRequestInfo import OrderRequestInfo


# constants
REJECT_NO_MEMORY = 2000
REJECT_DECODE_FAILURE = 2001
REJECT_UNKNOWN_FIXID = 2002
REJECT_UNKNOWN_ROID = 2003
REJECT_UNKNOWN_CLOID = 2004
REJECT_UNKNOWN_EVENT = 2005
REJECT_SE_FAILURE = 2006
REJECT_UPDATE_FAILURE = 2007
REJECT_FIX_FAILURE = 2008
REJECT_ENCODE_FAILURE = 2009
REJECT_SYSTEM_EXCEPTION = 2010
REJECT_USER_NOT_ACTIVE = 2011
REJECT_ROUTER_FAILURE = 2012
REJECT_INVALID_REQUEST = 2013
REJECT_DATABASE_ERROR = 2014
REJECT_COMMUNICATION_ERROR = 2015
REJECT_MISSSING_ORDERID = 2016
REJECT_FAILED_RISK_CHECK = 2017
REJECT_BAD_INBOUND_ORDER_HANDLING = 2018
REJECT_BUYING_POWER_EXCEEDED = 2019
REJECT_DESTINATION_DOWN = 2020
REJECT_FAILED_RISK_DUPLICATE_CHECK = 2021
REJECT_FAILED_RISK_ETB_CHECK = 2022
REJECT_FAILED_RISK_PENDING_ORDER_CHECK = 2023
REJECT_FAILED_RISK_AGGREGATE_EXECUTED_CHECK = 2024
REJECT_FAILED_RISK_SELF_TRADE_CHECK = 2025
REJECT_FAILED_RISK_ORDER_MARKING_CHECK = 2026
REJECT_FAILED_RISK_ERRONEOUS_PRICE_CHECK = 2027
REJECT_FAILED_RISK_INVALID_SIDE_CHECK = 2028
REJECT_FAILED_RISK_MAX_SHARES_CHECK = 2029
REJECT_FAILED_RISK_MAX_VALUE_CHECK = 2030
REJECT_FAILED_RISK_ON_CLOSE_CHECK = 2031
REJECT_CROSSING_ENGINE_REJECT = 2032
REJECT_POSITION_LIMIT_REACHED = 2033
REJECT_IPO_MKT_ORDER = 2034
REJECT_NOT_ENOUGH_LOCATES = 2035
REJECT_UNKNOWN_ACCOUNT = 2036
REJECT_MAX_LOSS_REACHED = 2037
REJECT_POSITION_OR_OPEN_LIMIT_REACHED = 2038
REJECT_MAX_ORDERS_PER_SECOND = 2039
REJECT_MAX_UNIQUE_POSITIONS = 2040
REJECT_MAX_EXECUTED_LIMIT = 2041
REJECT_MAX_OPEN_AND_EXECUTED_LIMIT = 2042
REJECT_ADV_PCT_EXCEEDED = 2043
REJECT_NYSE_TPA_NO_PARENT_ORDER = 2044
REJECT_FAILED_RISK_MAX_OPEN_ORDERS_CHECK = 2053
REJECT_FAILED_TICKSIZEPILOT_PRICE_CHECK = 2060
REJECT_CLIENT_POSITION_LIMIT_REACHED = 2061
REJECT_CLIENT_MAX_LOSS_REACHED = 2062
REJECT_CLIENT_MAX_OPEN_ORDERS_PER_SYMBOL = 2063
REJECT_EXCEEDS_UNROUTEDQTY = 2070
REJECT_MAX_SHARES_TRADED_LIMIT = 2076
REJECT_TRADING_NOT_ALLOWED = 2077
REJECT_CXL_REPLACE_CLOSED_ORDER = 2078
REJECT_USER_NOT_ALLOWED_TO_TRADE_SYMBOL = 2079
REJECT_NYSE_TPA_MISSING_BADGE_ID = 2080


class Reject(EzxMsg, metaclass=EzxMsgMeta):

	__block_fields__ = {
		'ORB' : Field('rejectedRequest', 'ORB', noop_decode_func, ), 
	}

	def __init__(self, **kwargs):
		self.msgType : int = None
		self.routerOrderID : int = None
		self.myID : str = None
		self.parentKey : str = None
		self.side : int = None
		self.symbol : str = None
		self.price : float = None
		self.orderQty : float = None
		self.returnCode : int = None
		self.returnDesc : str = None
		self.userName : str = None
		self.state : str = None
		self.leavesQty : float = None
		self.cumQty : float = None
		self.avgPrice : float = None
		self.unroutedQty : float = None
		self.rejectedRequest : OrderRequestInfo = None
		super(Reject, self).__init__(**kwargs)
		self.msg_type = api.IserverMsgType.APPLICATION.value
		self.msg_subtype = api.IserverMsgSubType.REJECT.value




'''
Created on 25 Mar 2022

@author: shalomshachne
'''
import unittest
from iserver.util import StopWatch
from time import sleep
from iserver import util


MAX_DELTA_MS = 30

MAX_DELTA_SECS = MAX_DELTA_MS / 1000

MAX_DELTA_NANOS = MAX_DELTA_MS * 1000000

MAX_DELTA_MICROS = MAX_DELTA_NANOS / 1000

class Test(unittest.TestCase):

    def assert_approximately_equals(self, value1, value2, delta=MAX_DELTA_MS):
        value1 = float(value1)
        value2 = float(value2)
        diff = abs(value1 - value2) 
        self.assertTrue(diff <= delta, f'expected diff<{delta}, actual diff={diff}. value={value1}, value2={value2}')

    def setUp(self):
        self.watch = StopWatch()


    def tearDown(self):
        pass

    def testInitialState(self):
        self.assertEquals(0, self.watch.elapsed_nanos())

    def testMeasuresElapsedTime(self):
        sleep_time = .200
        watch = self.watch
        watch.start()
        sleep(sleep_time)
        watch.stop()
        elapsed_nanos = watch.elapsed_nanos()
        self.assertTrue(elapsed_nanos > 0, 'tracked time')
        
        millis = watch.elapsed_millis()
        self.assert_approximately_equals(sleep_time * 1000, millis)
        
    def testMeasuresElapsedTimeSeconds(self):
        sleep_time = .200
        watch = self.watch
        watch.start()
        sleep(sleep_time)
        watch.stop()
        
        seconds = watch.elapsed_seconds()
        self.assert_approximately_equals(sleep_time, seconds, MAX_DELTA_SECS)
        
        
    def testElapsedDataTypes(self):
        sleep_time = .60
        watch = self.watch
        watch.start()
        sleep(sleep_time)
        watch.stop()
        
        value = watch.elapsed_nanos()
        self.assertIsInstance(value, int)        
        
        value = watch.elapsed_micros()
        self.assertIsInstance(value, float)           

        value = watch.elapsed_millis()
        self.assertIsInstance(value, float)      
        
        
        value = watch.elapsed_seconds()
        self.assertIsInstance(value, float)      
                
        
    def testMeasuresElapsedNanosAndMicros(self):
        sleep_time = .090
        watch = self.watch
        watch.start()
        sleep(sleep_time)
        watch.stop()
        
        expected = sleep_time * 1000000000
        nanos = watch.elapsed_nanos()
        self.assert_approximately_equals(expected, nanos, MAX_DELTA_NANOS)
        
        expected = sleep_time * 1000000
        micros = watch.elapsed_micros()
        self.assert_approximately_equals(expected, micros, MAX_DELTA_MICROS)
        
    
    def testStopNotPressedReturnsCurrentElapsed(self):
        sleep_time = .060
        watch = self.watch
        watch.start()
        
        sleep(sleep_time)
        
        expected = sleep_time * 1000
        actual = watch.elapsed_millis()
        self.assert_approximately_equals(expected, actual)
        
        sleep(sleep_time)
        expected *= 2
        actual = watch.elapsed_millis()
        self.assert_approximately_equals(expected, actual)
        
    def testReset(self):
        self.testMeasuresElapsedTime()
        self.watch.reset()
        self.assertEqual(0, self.watch.elapsed_nanos(), 'reset cleared times')
                
    
    def testGetRunningWatch(self):
        watch = util.start_timing()
        self.assertIsNotNone(watch)
        
        self.assertTrue(watch.is_running(), 'started running')
        
    def testIsRunning(self):
        self.assertFalse(self.watch.is_running())
        self.watch.start()
        self.assertTrue(self.watch.is_running())
        self.watch.stop()
        self.assertFalse(self.watch.is_running())
        
        
if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


'''
Created on 23 Feb 2022

@author: shalomshachne
'''
import unittest
from iserver.msgs.YieldData import YieldData
import iserver.enums
from iserver import ezx_msg


class Test(unittest.TestCase):

    def setUp(self):
        self.y = YieldData()

    def testYield(self):
        self.y.myYield = .023512
        self.y.yieldType = 'FIXED'
        
        encoded = ezx_msg.encode(self.y)
        y2 = YieldData()
        ezx_msg.populate_message(y2, encoded)
        self.assertEqual(self.y, y2)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testYield']
    unittest.main()



'''
 Generated by EZX Code Generator  
Date: 04/18/22 
'''


from iserver.EzxMsg import EzxMsg, EzxMsgMeta
from iserver.msgs import *
from iserver.enums import api

from .OrderStatus import OrderStatus
from .ParentOrderStatus import ParentOrderStatus
from .AllocationStatus import AllocationStatus



class GetOrdersResponse(EzxMsg, metaclass=EzxMsgMeta):

	__block_fields__ = {
		'OB' : Field('orderStatusList', 'OB', noop_decode_func, update_list_item), 
		'PB' : Field('parentOrderStatusList', 'PB', noop_decode_func, update_list_item), 
		'ALLOCSTATBLOCK' : Field('allocStatusList', 'ALLOCSTATBLOCK', noop_decode_func, update_list_item), 
	}

	def __init__(self, **kwargs):
		self.myID : str = None
		self.orderStatusList : list[OrderStatus] = None
		self.parentOrderStatusList : list[ParentOrderStatus] = None
		self.allocStatusList : list[AllocationStatus] = None
		super(GetOrdersResponse, self).__init__(**kwargs)
		self.msg_type = api.IserverMsgType.APPLICATION.value
		self.msg_subtype = api.IserverMsgSubType.GET_ORDERS_RESPONSE.value




'''
Created on 11 Mar 2022

@author: shalomshachne
'''
import inspect
import random
import string

from iserver.enums.msgenums import LogonType, Side, Events, MsgType, ReturnCode
from iserver.net import ConnectionInfo
from iserver.msgs.OrderResponse import OrderResponse
from iserver.msgs.OrderRequest import OrderRequest
from iserver.msgs.OrderRequestInfo import OrderRequestInfo
from iserver.msgs.LogonResponse import LogonResponse


__next_int = 0

def create_connection_info(host='localhost', port=15000, user='aUser', password='aPasswd!', company='MyCompany', logon_type=LogonType.ALL.value):
    info = inspect.getargvalues(inspect.currentframe())
    args = info.locals 
    return ConnectionInfo(**args)
    
def next_int_id():
    global __next_int
    __next_int = __next_int + 1
    return __next_int

def random_symbol(max_chars : int =4) -> str:
    return ''.join(random.choices(string.ascii_uppercase, k=max_chars))

def random_price(min_price : float = .01, max_price : float = 999.99) -> float:
    return round(random.uniform(min_price, max_price), 2)

def random_quantity(min_qty : int = 1, max_qty : int = 5000):
    return random.randint(min_qty, max_qty)

def create_order_response(side = Side.BUY.value, symbol = random_symbol(), orderQty = random_quantity(), price=random_price()):
    return OrderResponse(side=side, symbol=symbol, orderQty=orderQty, price=price, routerOrderID=random_quantity(), event=Events.ORDR.value)

def create_new_order_request(side = Side.BUY.value, symbol = random_symbol(), orderQty = random_quantity(), price=random_price()):
    return OrderRequest(side=side, symbol=symbol, orderQty=orderQty, price=price, routerOrderID=random_quantity(), msgType = MsgType.NEW.value)
 
def create_new_order_request_info(side = Side.BUY.value, symbol = random_symbol(), orderQty = random_quantity(), price=random_price()):
    return OrderRequestInfo(side=side, symbol=symbol, orderQty=orderQty, price=price, routerOrderID=random_quantity(), msgType = MsgType.NEW.value)

def create_logon_response(returnCode : int = ReturnCode.OK.value, returnDesc : str = None):
    return LogonResponse(returnCode = returnCode, returnDesc = returnDesc)
  


'''
 Generated by EZX Code Generator  
Date: 04/18/22 
'''


from iserver.EzxMsg import EzxMsg, EzxMsgMeta
from iserver.msgs import *
from iserver.enums import api




class QuoteRecord(EzxMsg, metaclass=EzxMsgMeta):


	def __init__(self, **kwargs):
		self.updateType : int = None
		self.quoteType : int = None
		self.symbol : str = None
		self.quoteSize : float = None
		self.price : float = None
		self.quoteID : str = None
		self.quoteSource : str = None
		self.quoteDateTime : str = None
		self.settleDate : str = None
		super(QuoteRecord, self).__init__(**kwargs)




'''
Created on 10 Mar 2022

@author: shalomshachne
'''

import logging
import socket
from socketserver import BaseRequestHandler, TCPServer
from threading import Condition
import threading

import unittest
from unittest.mock import Mock

from iserver import ezx_msg
from iserver.enums.api import IserverMsgSubType, IserverMsgType
from iserver.enums.msgenums import ReturnCode
from iserver.msgs.LogonRequest import LogonRequest
from iserver.msgs.LogonResponse import LogonResponse
from iserver.net import ConnectionInfo, ApiClient, ClientState, \
    NotLoggedInException
from iserver.util import next_id
from tests import check_logging, wait_for_condition, test_util, fill_array
from tests.data import test_data_factory
from iserver.EzxMsg import EzxMsg


logger = logging.getLogger(__name__)


class TestApiClient(unittest.TestCase):

    DEFAULT_PORT = 15002
        
    @classmethod
    def setUpClass(cls):
        # super(TestApiClient, cls).setUpClass()
        # sh = logging.StreamHandler(sys.stdout)
        # logging.root.addHandler(sh)
        check_logging()

    def start_server(self):
        self.server = MockIserver(TestApiClient.DEFAULT_PORT, self.handle)
        self.server.start()

    def setUp(self):
        self.connectInfo = ConnectionInfo(host='localhost', port=TestApiClient.DEFAULT_PORT, company='EZX', user='gnadan', password='gnadan', connect_retry_seconds=.1)        
        self.client = ApiClient(self.connectInfo)
        self.start_server()
        
        self.clientTest = lambda *args: None
        self.last_message = None  # message received by server from the ApiClient
        self.wait_condition = Condition()
        
        self.connect_wait_secs = 3
        
        self.incoming_messages = []
        
        
        
    def tearDown(self):
        self.client.stop()
        self.server.stop()        
    
    def on_state_change(self, state):
        self.last_state_change = state
        
    def handle(self, header, body):
        logger.info(f'got a message. header={header}, body={body}')
        self.wait_condition.acquire()
        self.last_message = (header, body)
        self.wait_condition.notify_all()
        self.wait_condition.release()
            
    def client_message_handler(self, subtype : int, msg : EzxMsg):
        self.incoming_messages.append(msg)
        
    def decode_last_message(self):
        if self.last_message:
            header, body = self.last_message
            _, _, subtype = ezx_msg.parse_header(header)
            body = body.decode('UTF-8')
            return ezx_msg.decode_message(subtype, body)
    

    def testNotifiesStateChange(self):
        received = 'untouched'
        
        def printState(state): 
            nonlocal received        
            received = state
            logger.debug('got state change=%s', received)
        
        handler = printState
                   
        self.client.on_state_change = handler
         
        self.client._set_state(ClientState.CONNECTED)
        
        self.assertEqual(ClientState.CONNECTED, received, "called back on state change")
    
    def test_state_functions(self):
        self.assertFalse(self.client.is_connected(), 'valid initial state')
        
        self.client._set_state(ClientState.CONNECTED)
        self.assertTrue(self.client.is_connected())
        
        self.client._set_state(ClientState.LOGGED_IN)
        self.assertTrue(self.client.is_connected())

    def assert_logon_received(self):
        message = self.last_message
        self.assertIsNotNone(message, 'got a message')
        header, body = message
        sub_type = int(header[8:11].decode('UTF-8'))
        self.assertEqual(IserverMsgSubType.LOGON.value, sub_type, 'received logonRequest')
        logon = ezx_msg.decode_message(sub_type, body.decode('UTF-8'))
        self.assertEqual(0, logon.seqNo, 'default logon seqNo = 0')

    def test_start_sends_logon(self):
        self.wait_condition.acquire()
        self.client.start()
        self.wait_condition.wait(self.connect_wait_secs)  # 3 seconds max wait.

        self.assert_logon_received()
        
    def test_can_restart_after_stop(self):
        self.test_start_sends_logon()
        self.client.stop()
        self.assertEqual(ClientState.STOPPED, self.client.state(), 'valid initial state')
        self.last_message = None

        wait_for_condition(lambda: self.client.connection_monitor.monitor_thread == None, 2)
        self.assertIsNone(self.client.connection_monitor.monitor_thread, 'valid test state - monitor is fully shut down')

        print('===>>>> restarting client here....')                
        self.test_start_sends_logon()  # restart and show it can be reused
        self.assertTrue(self.client.is_running(), 'started...')
        self.assertEqual(ClientState.CONNECTED, self.client.state())
        
    
    def test_can_set_logon_seqno(self):
        self.client.last_seqno = 10
        self.wait_condition.acquire()
        self.client.start()
        self.wait_condition.wait(self.connect_wait_secs)  # 3 seconds max wait.

        logon = self.decode_last_message()
        self.assertEqual(10, logon.seqNo, 'used last seqNo set on the client')
        
    def test_notifies_logon_response(self):
        self.client.on_state_change = self.on_state_change
                
        response = LogonResponse(returnCode=ReturnCode.OK.value)
        api_bytes = ezx_msg.to_api_bytes(response, test_data_factory.random_quantity())
        self.client._handle_message(response.msg_type, response.msg_subtype, api_bytes[15:])
        
        self.assertEqual(ClientState.LOGGED_IN, self.client.state(), 'handled logonresponse correctly = ok')
        self.assertEqual(ClientState.LOGGED_IN, self.last_state_change, 'got notified on logon')
        
    def test_logon_failure_shuts_down_client(self):
        counter = 0

        def check_logon_failure_notification(state):
            nonlocal counter
            counter = counter + 1
            if counter == 1:
                self.assertEqual(ClientState.LOGON_FAILURE, state, 'notified of logon failure')
                self.got_state_change = True

        self.client.on_state_change = check_logon_failure_notification
        failure = 'wrong user name!'
        response = LogonResponse(returnCode=ReturnCode.INVALID_USER.value, returnDesc=failure)
        api_bytes = ezx_msg.to_api_bytes(response, test_data_factory.random_quantity())

        self.client._handle_message(response.msg_type, response.msg_subtype, api_bytes[15:])
        
        self.assertEqual(ClientState.STOPPED, self.client.state(), 'handled logonresponse correctly = failure')
        self.assertTrue(self.client.is_stopped(), 'shut down on logon failure')
        self.assertTrue(self.got_state_change)
        
        errcode, desc = self.client.get_logon_failure()
        self.assertEqual(ReturnCode.INVALID_USER.value, errcode)
        self.assertEqual(failure, desc)
                        
    def test_client_retries_connection(self):
        self.test_start_sends_logon()
        logger.info('stopping server to cause ApiClient disconnect')
        self.server.stop()
        self.last_message = None
        self.connect_wait_secs = 10  # not sure how long to wait.
        self.wait_condition.acquire()
        logger.info('restarting server...')        
        self.start_server()
        logger.info('waiting for login...')        
        self.wait_condition.wait(self.connect_wait_secs) 
        
        # should reconnect shortly
        # message = self.last_message
        # self.assertIsNotNone(message, 'reconnected and logged in')
        self.assert_logon_received()          
    
    def test_client_can_start_before_server(self):
        self.server.stop()
        
        wait_for_condition(lambda: not self.server.is_running())
        self.assertFalse(self.server.is_running(), 'valid test state, server is shut down')                
        self.client.start()

        self.wait_condition.acquire()        
        self.start_server()
        self.wait_condition.wait(self.connect_wait_secs)         
        self.assert_logon_received()
 
    def test_write_not_connected(self):
        response = test_data_factory.create_order_response()
        with self.assertRaises(NotLoggedInException):
            self.client.send_message(response)
            
    def test_write_connected_can_send_logon(self):
        self.client._set_state(ClientState.CONNECTED)
        mock_socket = Mock()
        mock_socket.sendall(bytearray())
        self.client._socket = mock_socket
        
        mock_socket.sendall.assert_called_once()
        logon = LogonRequest()
        self.client.send_message(logon)
    
    def test_calling_start_twice_does_nothing(self):
        mock_monitor = Mock()
        self.client.connection_monitor = mock_monitor
        mock_monitor.start()
        
        self.client.start()
        mock_monitor.start.assert_called_once()
 



    def test_catches_handler_exception(self):

        # if user code throws an Exception, client should keep processing.
        def exception_thrower(subtype, msg):
            raise Exception('something bad happened (unit test)')
        
        self.client._msg_handler = exception_thrower
        msg = test_data_factory.create_order_response()
        
        self.push_message_to_client(msg) 
        
    def test_updates_seqno(self):
        response = test_data_factory.create_order_response()
        
        response.seqNo = test_data_factory.next_int_id()
        self.push_message_to_client(response)        
        self.assertEqual(response.seqNo, self.client.last_seqno, 'updated seqno')
        
        response.seqNo = response.seqNo + 1
        self.push_message_to_client(response)
        self.assertEqual(response.seqNo, self.client.last_seqno, 'updated seqno')
        
    
    def test_cannot_set_seqno_to_bad_values(self):
        seqNo = test_data_factory.next_int_id()
        self.client.last_seqno = seqNo
        self.client.last_seqno = None
        
        self.assertEqual(seqNo, self.client.last_seqno, 'cannot set to None')
        
        self.client.last_seqno = seqNo - 1
        self.assertEqual(seqNo, self.client.last_seqno, 'cannot decrement seqNo')
        
    def test_send_heartbeat(self):
        s = MockSocket()
        self.client._socket = s
        self.client._set_state(ClientState.LOGGED_IN)
        self.client._heartbeat()
    
    
        sent_msgs = s.received
        self.assertEqual(1, len(sent_msgs), 'sent a message')
        msg_bytes = sent_msgs[0]
        self.assertEqual(16, len(msg_bytes), 'header + etx')
        self.assertEqual(3, msg_bytes[15])
        
        length, mtype, subtype = ezx_msg.parse_header(msg_bytes)
        self.assertEqual(IserverMsgType.APPLICATION.value, mtype)
        self.assertEqual(IserverMsgSubType.HEARTBEAT.value, subtype, 'expected msubtype') 
        self.assertEqual(1, length, 'heartbeat has no msg contents except ETX character')
        
          
        
    
    
    def test_mock_socket(self):
        s = MockSocket()
        msg_bytes = bytearray()
        s.sendall(msg_bytes)
        
        self.assertEqual(1, len(s.received))
        self.assertEqual(msg_bytes, s.received[0])
        
    def test_socket_receive_partial_body_reads_continues_to_read(self):
        header_string = '00035702150028'
        body_string = 'MYID=Ord-637878923112570755-22ROID=78UIDN=7UNAM=igorSIDE=2PX=0.02OQTY=500MINQ=100FV=0DISOFF=0XSYM=UNYSYM=ULXMDSYM=UNYMDSYM=ULXTXT=+B.1.a.4YTXT=+B.1.a.4DEST=SIMUACCT=5EW6PK15LREQD=GOLDMANHGALL=100MVALL=0SLIP=0.01ISCAN=10HSCAN=15REPMS=0MAXDIST=40DELAYMS=0FQTY=0HQTY=0XFQTY=0YFQTY=0PRSTATE=CanceledSTATE=CANDOID=pairs-server\x003'
        self.assertEqual(int(header_string[0:6]),len(body_string), 'valid setup')
        
        self.client._msg_handler = self.client_message_handler 
        
        
        header_bytes = header_string.encode('UTF8')
        body_bytes = body_string.encode('UTF-8')
        partial_recv_length = 213
        
        counter = -1
        values = [15, partial_recv_length, 0, len(body_bytes) - partial_recv_length]
        def recv_into(b):
            import tests
            nonlocal counter            
            counter += 1
            if counter == 0:
                b[:] = header_bytes
            elif counter == 1:
                tests.fill_array(b, body_bytes[0 : partial_recv_length])
            elif counter == 2:
                pass
            elif counter == 3:
                tests.fill_array(b, body_bytes[partial_recv_length:])
            else:
                raise ConnectionAbortedError #stop processing
            return values[counter]
        
        s = Mock()
        s.recv_into = recv_into
        enter_s = Mock(return_value=s)
        s.__enter__ = enter_s
        s.__exit__ = enter_s
        self.client._socket = s
        self.client._recv()
        
        expected = 1
        actual = len(self.incoming_messages)
        self.assertEqual(expected, actual, 'got correct number of messages')
        
    def test_socket_receive_partial_header(self):
        header_string = '00035702150028'
        body_string = 'MYID=Ord-637878923112570755-22ROID=78UIDN=7UNAM=igorSIDE=2PX=0.02OQTY=500MINQ=100FV=0DISOFF=0XSYM=UNYSYM=ULXMDSYM=UNYMDSYM=ULXTXT=+B.1.a.4YTXT=+B.1.a.4DEST=SIMUACCT=5EW6PK15LREQD=GOLDMANHGALL=100MVALL=0SLIP=0.01ISCAN=10HSCAN=15REPMS=0MAXDIST=40DELAYMS=0FQTY=0HQTY=0XFQTY=0YFQTY=0PRSTATE=CanceledSTATE=CANDOID=pairs-server\x003'
        
        self.client._msg_handler = self.client_message_handler 
        header_bytes = header_string.encode('UTF8')
        body_bytes = body_string.encode('UTF-8')

        partial_header = 2
        s = MockSocket()
        s.to_send.append(header_bytes[0:partial_header]) #send only first part of the header
        s.to_send.append(header_bytes[partial_header:])  #send the rest
        s.to_send.append(body_bytes) #body 
                
        self.client._socket = s
        self.client._recv()
        
        expected = 1
        actual = len(self.incoming_messages)
        self.assertEqual(expected, actual, 'got correct number of messages')

    def test_socket_receives_multiple_messages(self):
        self.client._msg_handler = self.client_message_handler
        msgs = []
        msgs.append(ezx_msg.to_api_bytes(test_data_factory.create_order_response()))
        msgs.append(ezx_msg.to_api_bytes(test_data_factory.create_new_order_request()))
        msgs.append(ezx_msg.to_api_bytes(test_data_factory.create_logon_response()))
        
        s = MockSocket(to_send=msgs)        
        
        self.client._socket = s
        self.client._recv()
        expected = len(msgs)
        actual = len(self.incoming_messages)
        message = "client received all the messages"
        self.assertEqual(expected, actual, message)
        
        
        
        
        







        
    def push_message_to_client(self, msg):
        body = ezx_msg.encode(msg).encode(encoding='UTF-8')
        self.client._handle_message(IserverMsgType.APPLICATION.value, IserverMsgSubType.ORDERRESPONSE.value, body)


        
    
class MockIserver(object):

    def __init__(self, port, handler):
        
        class ApiClientTestHandler(BaseRequestHandler):

            def __init__(self, *args):
                self.handler = handler  # function(header,body)
                super().__init__(*args)
                logger.debug('ApiClientTestHandler..__init__(): initialized handler')
                
            def handle(self):
                # parse header
                header = self.request.recv(15)
                if not header:
                    logger.error('error: did not receive enough bytes for an API header!')
                    return 
                length, msg_type, msg_subtype = ezx_msg.parse_header(header)
                logger.info(f'msg length={length}, msg_subtype={msg_subtype}')
                body = self.request.recv(length)
                self.handler(header, body)
                 
        self.port = port
        self.server = TCPServer(('localhost', port), ApiClientTestHandler, False)
        # enable immediate reuse
        self.server.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server.server_bind()
        self.server.server_activate()
        logger.info(f'Server listening on: {self.server.server_address}')
        
    def start(self):
        logger.info(f'starting server on port={self.port}')
        t = threading.Thread(target=self.server.serve_forever)
        t.daemon = True
        t.start()
        self.server_thread = t
        logger.info(f'Server started.')
        # self.server.serve_forever()
        
    def is_running(self):
        try:
            return self.server_thread.is_alive()
        except AttributeError:
            pass
        
    def stop(self):
        logger.info('stopping server')
        self.server.shutdown()
        self.server.server_close()
        logger.info('server stopped')
        

class MockSocket(socket.socket):
    def __init__(self, to_send : list = list()):
        super().__init__(socket.AF_INET, socket.SOCK_STREAM)
        self.received = []
        self.to_send : list(bytearray) = to_send
        self.message_send_index = 0
        self.within_message_index = 0
        
    def sendall(self, *args, **kwargs):
        msg_bytes = args[0]
        self.received.append(msg_bytes)        
        return len(msg_bytes)
    
    def recv_into(self, buffer, nbytes=0, flags=None) -> int:
        if self.message_send_index < len(self.to_send):
            source = self.to_send[self.message_send_index][self.within_message_index:]
            sent = fill_array(buffer, source)
            if sent == len(source): # fully sent
                self.message_send_index += 1
                self.within_message_index = 0
            else:   # partial message sent
                self.within_message_index = sent
                
            return sent
        
        raise ConnectionAbortedError # when no more messages

if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



'''
Created on Sep 29, 2021

@author: Sgoldberg
'''

import unittest

from iserver.msgs import LogonRequest


SAMPLE_MSG = "00010402100000\001CONAM=FEIS\001DTGMT=<TODAY>\001TYPE=3\001PWD=igor\001SEQ=0\001UNAM=igor\001SNDDESTSTRAT=1\001SNDSETS=1\001DSTAT=1\001OMSVER=4.0.1\001\003"

class Test(unittest.TestCase):

    def setUp(self):
        self.logon = LogonRequest()


    def tearDown(self):
        pass


    def testConstructorWithSeqNo(self):
        seqno = 101
        l = LogonRequest(seqNo = seqno)
        self.assertEqual(seqno, l.seqNo)


    def testEncode(self):
        expected = SAMPLE_MSG
        c = expected[14]
        self.assertEqual(1, ord(c), "ASCII 1 correct")
        c = expected[-1]
        self.assertEqual(3, ord(c), "ASCII 3 correct")
        
        logon = self.logon
        logon.companyName = 'FEIS'                
        logon.user = 'igor'
        logon.password = 'igor'
        logon.date = '<TODAY>'
        logon.logonType = 3
        logon.sendDestStrategies = 1
        logon.sendSettings = 1
        logon.sendDestinationState = 1
        logon.omsVersion = "4.0.1"
        
        msg_only = expected[15:-1]
        print('msg=' + msg_only)
        
    def testMsgTypeAndSubType(self):
        logon = self.logon
        self.assertEqual(2, logon.msg_type, "correct msgtype")
        self.assertEqual(100, logon.msg_subtype, "correct msgtype")
        


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


'''
Created on 29 Apr 2022

@author: shalomshachne
'''
import unittest
import argparse



class Test(unittest.TestCase):


    def testParserExit(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('side', type=str)
        parser.add_argument('symbol', type=str)
        for action in parser._actions: action.required = False
        args = ['B']
        
        r = parser.parse_args(args)
        self.assertIsNotNone(r)
        
        


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testParserExit']
    unittest.main()


'''
Created on 21 Apr 2022

@author: shalomshachne
'''
import unittest

from iserver import util
from iserver.enums.api import IserverMsgSubType
from iserver.enums.msgenums import MsgType, Side, OrdType
from iserver.msgs.convenience_msgs import ReplaceOrder, CancelOrder, NewOrder
from tests.data import test_data_factory


class Test(unittest.TestCase):

    def testReplaceOrderStandardConstructor(self):
        orderId = test_data_factory.next_int_id()
        price = test_data_factory.random_price()
        qty = test_data_factory.random_quantity()
        
        replace = ReplaceOrder(orderId, price, qty)
        
        field = "api msg_subtype"
        self.assertEqual(IserverMsgSubType.ORDER.value, replace.msg_subtype, field)
                
        field = "msgType"
        self.assertEqual(MsgType.REPL.value, replace.msgType, field)
        
        field = "routerOrderID"
        self.assertEqual(orderId, replace.routerOrderID, field)
        
        field = "price"
        self.assertEqual(price, replace.price, field)
        
        field = "qty"
        self.assertEqual(qty, replace.orderQty, field)
    
    def testReplaceKwargsWorks(self):
        orderId = test_data_factory.next_int_id()
        account = util.next_id()
        args = {'account': account}
        replace = ReplaceOrder(orderId, **args)
        
        field = "api msg_subtype"
        self.assertEqual(IserverMsgSubType.ORDER.value, replace.msg_subtype, field)

        field = "routerOrderID"
        self.assertEqual(orderId, replace.routerOrderID, field)
        
        field = "account set from kwargs"
        self.assertEqual(account, replace.account, field)

    def testReplaceKwargsWorks2(self):
        orderId = test_data_factory.next_int_id()
        account = util.next_id()
        
        replace = ReplaceOrder(orderId, account=account)
        
        field = "api msg_subtype"
        self.assertEqual(IserverMsgSubType.ORDER.value, replace.msg_subtype, field)

        field = "routerOrderID"
        self.assertEqual(orderId, replace.routerOrderID, field)
        
        field = "account set from kwargs"
        self.assertEqual(account, replace.account, field)
    
    def testCancel(self):
        orderId = test_data_factory.next_int_id()
        cancel = CancelOrder(orderId)
        
        field = "api msg_subtype"
        self.assertEqual(IserverMsgSubType.ORDER.value, cancel.msg_subtype, field)
 
        field = "msgType"
        self.assertEqual(MsgType.CANC.value, cancel.msgType, field)
 
        field = "routerOrderID"
        self.assertEqual(orderId, cancel.routerOrderID, field)


    def testNewOrderWithBasicFields(self):
        #     order = OrderRequest(msgType=MsgType.NEW.value, symbol=symbol, side=parse_side(side), orderQty=qty, price=price, ordType=ordType, \
        # destination=destination, myID=util.next_id())
        
        symbol = test_data_factory.random_symbol()
        side = Side.SELL_PLUS.value
        price = test_data_factory.random_price()
        qty = test_data_factory.random_quantity()
        ordType = OrdType.LIMIT_ON_CLOSE.value
        destination = 'SIMU'
        id = util.next_id
                          
        order = NewOrder(symbol, side, qty, price, destination, id, ordType)
        
        field = "api msg_subtype"
        self.assertEqual(IserverMsgSubType.ORDER.value, order.msg_subtype, field)        
        
        field = "msgType"
        self.assertEqual(MsgType.NEW.value, order.msgType, field)       
        
        field = "symbol"
        self.assertEqual(symbol, order.symbol, field)       
        
        field = "side"
        self.assertEqual(side, order.side, field)       

        field = "qty"
        self.assertEqual(qty, order.orderQty, field)       

        field = "ordType"
        self.assertEqual(ordType, order.ordType, field)       

        field = "destination"
        self.assertEqual(destination, order.destination, field)       

        field = "myID"
        self.assertEqual(id, order.myID, field)       

        field = "symbol"
        self.assertEqual(symbol, order.symbol, field)       
        
        
    def testNewOrderWithKwargs(self):
        #     order = OrderRequest(msgType=MsgType.NEW.value, symbol=symbol, side=parse_side(side), orderQty=qty, price=price, ordType=ordType, \
        # destination=destination, myID=util.next_id())
        
        symbol = test_data_factory.random_symbol()
        side = Side.SELL_PLUS.value
        price = test_data_factory.random_price()
        qty = test_data_factory.random_quantity()
        ordType = OrdType.LIMIT_ON_CLOSE.value
        destination = 'SIMU'
        id = util.next_id
        
        account = 'acc' + str(test_data_factory.next_int_id())
                          
        order = NewOrder(symbol, side, qty, price, destination, id, ordType, account=account)
        
        field = "msgType"
        self.assertEqual(MsgType.NEW.value, order.msgType, field)               
        
        self.assertEqual(account, order.account, 'account set')
        
        
    def testNewOrderDefaultsOrdTypeToLimit(self):
        #     order = OrderRequest(msgType=MsgType.NEW.value, symbol=symbol, side=parse_side(side), orderQty=qty, price=price, ordType=ordType, \
        # destination=destination, myID=util.next_id())
        
        symbol = test_data_factory.random_symbol()
        side = Side.SELL_PLUS.value
        price = test_data_factory.random_price()
        qty = test_data_factory.random_quantity()
        destination = 'SIMU'
        id = util.next_id
        
        account = 'acc' + str(test_data_factory.next_int_id())
                          
        order = NewOrder(symbol, side, qty, price, destination, id, account=account)
        
        field = "msgType"
        self.assertEqual(MsgType.NEW.value, order.msgType, field)               
        
        field = 'account set'
        self.assertEqual(account, order.account, field)
                    
        field = 'defaulted ordType Limit'
        self.assertEqual(OrdType.LIMIT.value, order.ordType, field)
        
    def testNewOrderDefaultsOrdTypeToMarket(self):
        #     order = OrderRequest(msgType=MsgType.NEW.value, symbol=symbol, side=parse_side(side), orderQty=qty, price=price, ordType=ordType, \
        # destination=destination, myID=util.next_id())
        
        symbol = test_data_factory.random_symbol()
        side = Side.SELL_PLUS.value
        qty = test_data_factory.random_quantity()
                                  
        order = NewOrder(symbol, side, qty)
                            
        field = 'defaulted ordType Market'
        self.assertEqual(OrdType.MARKET.value, order.ordType, field)   
        
    def testNewOrderDefaultsOrdTypeToMarketPriceZero(self):
        #     order = OrderRequest(msgType=MsgType.NEW.value, symbol=symbol, side=parse_side(side), orderQty=qty, price=price, ordType=ordType, \
        # destination=destination, myID=util.next_id())
        
        symbol = test_data_factory.random_symbol()
        side = Side.SELL_PLUS.value
        qty = test_data_factory.random_quantity()
        
                         
        order = NewOrder(symbol, side, qty, 0)
                            
        field = 'defaulted ordType Market'
        self.assertEqual(OrdType.MARKET.value, order.ordType, field)                
        self.assertIsNone(order.price, 'blanked out price')


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testReplaceOrder']
    unittest.main()



'''
Created on 18 Apr 2022

@author: shalomshachne
'''
import unittest

from iserver.msgs import Reject
from tests.data import test_data_factory
from iserver import util, ezx_msg
from pip._internal import req


class Test(unittest.TestCase):


    def setUp(self):
        self.reject = Reject.Reject()


    def tearDown(self):
        pass


    def testDecodeRejectWithOrderRequest(self):
        req = test_data_factory.create_new_order_request_info()
        req.myID = util.next_id()
        
        reject = self.reject
        reject.msgType = req.msgType
        reject.myID = req.myID
        reject.rejectedRequest = req
        reject.returnCode = Reject.REJECT_BUYING_POWER_EXCEEDED
        reject.returnDesc = "REJECT_BUYING_POWER_EXCEEDED"
        
        encoded = ezx_msg.encode(reject)
        decoded = ezx_msg.decode_message(reject.msg_subtype, encoded)
        
        self.assertEqual(reject.returnCode, decoded.returnCode, "returnCode")
        self.assertEquals(Reject.REJECT_BUYING_POWER_EXCEEDED, decoded.returnCode)
        self.assertEquals(reject.returnDesc, decoded.returnDesc)
        
        self.assertEqual(req, decoded.rejectedRequest, 'decoded the rejected request')
        
        
        
        
        


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


'''
 Generated by EZX Code Generator  
Date: 04/18/22 
'''


from iserver.EzxMsg import EzxMsg, EzxMsgMeta
from iserver.msgs import *
from iserver.enums import api

from .ChoiceGroup import ChoiceGroup
from .UpDown import UpDown
from .OfficeGroupClient import OfficeGroupClient
from .TraderList import TraderList



class Settings(EzxMsg, metaclass=EzxMsgMeta):

	__block_fields__ = {
		'CHOGBLK' : Field('choiceGroups', 'CHOGBLK', noop_decode_func, update_list_item), 
		'UPDWNBLK' : Field('updowns', 'UPDWNBLK', noop_decode_func, update_list_item), 
		'OGCBLK' : Field('officeGroupClients', 'OGCBLK', noop_decode_func, update_list_item), 
		'TRLST' : Field('tradersList', 'TRLST', noop_decode_func, update_list_item), 
	}

	def __init__(self, **kwargs):
		self.state : str = None
		self.gmtEndOfDay : str = None
		self.choiceGroups : list[ChoiceGroup] = None
		self.updowns : list[UpDown] = None
		self.officeGroupClients : list[OfficeGroupClient] = None
		self.tradersList : list[TraderList] = None
		super(Settings, self).__init__(**kwargs)




'''
Created on Sep 29, 2021

@author: Sgoldberg
'''
import unittest

from iserver.msgs import msgtags
from iserver.msgs.LogonResponse import LogonResponse
from iserver.msgs.OrderResponse import OrderResponse
from iserver.msgs import Field, update_list_item
from tests import random_string, random_number


class Test(unittest.TestCase):


    def testMsgTags(self):
        f = msgtags.get_api_field('companyName')
        self.assertIsNotNone(f, "got companyName field")

    def test_field_decode_string(self):
        f = Field('field1', 'FNAM')
        value = random_string()
        result = f.decode_value(value)
        self.assertEquals(value, result)
        
    def test_field_decode_int(self):
        f = Field('logonType', 'TYPE', int)
        value = str(random_number(0, 10))
        result = f.decode_value(value)
        self.assertEquals(int(value), result)
        
    def test_field_update_string(self):
        m = LogonResponse()
        f = Field('returnCode', 'RETC', int)
        value = "0"
        f.update_msg_field(m, value)
        self.assertEqual(0, m.returnCode)
        
    def test_field_update_list_attribute(self):
        m = LogonResponse()
        self.assertIsNone(m.destinations)
        f = Field('destinations', 'DEST', update_func = update_list_item)
        value = "SIMU1"
        f.update_msg_field(m, value)
        self.assertEqual(1, len(m.destinations))
        self.assertEquals(value, m.destinations[0])
        
    def test_name_collisions(self):
        self.assertTrue('DEST' in msgtags._api_tag_collisions)
        fields = msgtags._api_tag_collisions.get('DEST')
        self.assertEqual(2, len(fields))
        f = fields[0]
        self.assertEqual('destinations', f.name)
        f = fields[1]
        self.assertEqual('destination', f.name)
    
    def test_get_api_field_which_has_collision(self):
        api_tag = 'DEST'
        mo = LogonResponse()
        f = msgtags.get_msg_field(api_tag, mo)
        self.assertIsNotNone(f)
        self.assertEqual('destinations', f.name)

        mo = OrderResponse()
        f = msgtags.get_msg_field(api_tag, mo)
        self.assertIsNotNone(f)
        self.assertEqual('destination', f.name, "got the OrderResponse field definition")

        


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


'''
 Generated by EZX Code Generator  
Date: 04/18/22 
'''


from iserver.EzxMsg import EzxMsg, EzxMsgMeta
from iserver.msgs import *
from iserver.enums import api

from .Properties import Properties



class CompanySettingsUpdate(EzxMsg, metaclass=EzxMsgMeta):

	__block_fields__ = {
		'PRPSBLK' : Field('settings', 'PRPSBLK', noop_decode_func, ), 
	}

	def __init__(self, **kwargs):
		self.settings : Properties = None
		self.notes : str = None
		super(CompanySettingsUpdate, self).__init__(**kwargs)
		self.msg_type = api.IserverMsgType.APPLICATION.value
		self.msg_subtype = api.IserverMsgSubType.COMPANY_SETTINGS_UPDATE.value




'''
 Generated by EZX Code Generator  
Date: 04/18/22 
'''


from iserver.EzxMsg import EzxMsg, EzxMsgMeta
from iserver.msgs import *
from iserver.enums import api
from iserver.msgs.ClientOrderStatus import ClientOrderStatus

from .StrategyInfo import StrategyInfo
from .ExecutionStatus import ExecutionStatus



class ClientOrderResponse(ClientOrderStatus, metaclass=EzxMsgMeta):

	__block_fields__ = {
		'STGBLK' : Field('strategyInfo', 'STGBLK', noop_decode_func, ), 
		'EB' : Field('execStatus', 'EB', noop_decode_func, ), 
	}

	def __init__(self, **kwargs):
		self.strategyInfo : StrategyInfo = None
		self.execStatus : ExecutionStatus = None
		super(ClientOrderResponse, self).__init__(**kwargs)
		self.msg_type = api.IserverMsgType.APPLICATION.value
		self.msg_subtype = api.IserverMsgSubType.CLIENT_ORDER_RESPONSE.value




'''
 Generated by EZX Code Generator  
Date: 04/18/22 
'''

from  iserver.enums.api import IserverMsgSubType


from .ForgotPasswordRequest import ForgotPasswordRequest
from .OrderInfoResponse import OrderInfoResponse
from .ClientOrderResponse import ClientOrderResponse
from .BroadcastResponse import BroadcastResponse
from .UpdatePosition import UpdatePosition
from .CreditUsedUpdate import CreditUsedUpdate
from .QuoteAccept import QuoteAccept
from .ParentOrderRequest import ParentOrderRequest
from .AllocationResponse import AllocationResponse
from .AllocationInfoRequest import AllocationInfoRequest
from .DestinationStatus import DestinationStatus
from .Reject import Reject
from .CreateGroupAccount import CreateGroupAccount
from .QuoteResponse import QuoteResponse
from .GetOrdersResponse import GetOrdersResponse
from .PositionResponse import PositionResponse
from .MarketDataSnapshot import MarketDataSnapshot
from .OrderRequestList import OrderRequestList
from .LogonResponse import LogonResponse
from .OrderRequest import OrderRequest
from .ExecutionResponse import ExecutionResponse
from .DeleteClientOrder import DeleteClientOrder
from .HeartBeatMsg import HeartBeatMsg
from .OrderInfoRequest import OrderInfoRequest
from .AdminResponse import AdminResponse
from .MarketDataIncremental import MarketDataIncremental
from .UpdateGroupSettings import UpdateGroupSettings
from .ChangePasswordResponse import ChangePasswordResponse
from .MarketDataHealthMessage import MarketDataHealthMessage
from .ReadGroupAccounts import ReadGroupAccounts
from .UpdateGroupName import UpdateGroupName
from .FixedIncomeMarketDataIncrementalRefresh import FixedIncomeMarketDataIncrementalRefresh
from .AllocationInfoResponse import AllocationInfoResponse
from .ExecutionRequest import ExecutionRequest
from .ClientSettingsUpdate import ClientSettingsUpdate
from .DeleteGroup import DeleteGroup
from .OMSSettings import OMSSettings
from .ChangePasswordRequest import ChangePasswordRequest
from .GetOrdersRequest import GetOrdersRequest
from .Broadcast import Broadcast
from .Logout import Logout
from .DeleteAccount import DeleteAccount
from .PairsSpreadOrderResponse import PairsSpreadOrderResponse
from .QuoteRequest import QuoteRequest
from .TestChildClass import TestChildClass
from .IOI import IOI
from .ClientOrderRequest import ClientOrderRequest
from .LogonRequest import LogonRequest
from .TestParentClass import TestParentClass
from .ParentOrderResponse import ParentOrderResponse
from .OrderResponse import OrderResponse
from .ForgotPasswordResponse import ForgotPasswordResponse
from .Quote import Quote
from .MoveAccount import MoveAccount
from .CompanySettingsUpdate import CompanySettingsUpdate
from .EngineStatus import EngineStatus
from .AllocationRequest import AllocationRequest
from .AdminCommand import AdminCommand
from .CancelAll import CancelAll
from .PairsSpreadOrder import PairsSpreadOrder
from .TagValueMsg import TagValueMsg

msgs_by_subtype = {
	IserverMsgSubType.FORGOTPASSWORD.value: ForgotPasswordRequest,
	IserverMsgSubType.ORDERINFO_RESPONSE.value: OrderInfoResponse,
	IserverMsgSubType.CLIENT_ORDER_RESPONSE.value: ClientOrderResponse,
	IserverMsgSubType.BROADCAST_RESPONSE.value: BroadcastResponse,
	IserverMsgSubType.UPDATE_POSITION.value: UpdatePosition,
	IserverMsgSubType.CREDIT_USED_UPDATE.value: CreditUsedUpdate,
	IserverMsgSubType.QUOTE_ACCEPT.value: QuoteAccept,
	IserverMsgSubType.PARENT_ORDER_REQUEST.value: ParentOrderRequest,
	IserverMsgSubType.ALLOCATION_RESPONSE.value: AllocationResponse,
	IserverMsgSubType.ALLOCINFO_REQUEST.value: AllocationInfoRequest,
	IserverMsgSubType.DESTINATION_STATUS.value: DestinationStatus,
	IserverMsgSubType.REJECT.value: Reject,
	IserverMsgSubType.CREATE_GROUP_ACCOUNT.value: CreateGroupAccount,
	IserverMsgSubType.QUOTE_RESPONSE.value: QuoteResponse,
	IserverMsgSubType.GET_ORDERS_RESPONSE.value: GetOrdersResponse,
	IserverMsgSubType.POSITION_RESPONSE.value: PositionResponse,
	IserverMsgSubType.MARKETDATASNAPSHOT.value: MarketDataSnapshot,
	IserverMsgSubType.ORDERREQUESTLIST.value: OrderRequestList,
	IserverMsgSubType.LOGONRESPONSE.value: LogonResponse,
	IserverMsgSubType.ORDER.value: OrderRequest,
	IserverMsgSubType.EXECUTION_RESPONSE.value: ExecutionResponse,
	IserverMsgSubType.DELETE_CLIENT_ORDER.value: DeleteClientOrder,
	IserverMsgSubType.HEARTBEAT.value: HeartBeatMsg,
	IserverMsgSubType.ORDERINFO_REQUEST.value: OrderInfoRequest,
	IserverMsgSubType.ADMIN_RESPONSE.value: AdminResponse,
	IserverMsgSubType.MARKETDATAINCREMENTAL.value: MarketDataIncremental,
	IserverMsgSubType.UPDATE_GROUP_SETTINGS.value: UpdateGroupSettings,
	IserverMsgSubType.CHANGEPASSWORD_RESPONSE.value: ChangePasswordResponse,
	IserverMsgSubType.MARKETDATA_HEALTH.value: MarketDataHealthMessage,
	IserverMsgSubType.READ_GROUP_ACCOUNTS.value: ReadGroupAccounts,
	IserverMsgSubType.UPDATE_GROUP_NAME.value: UpdateGroupName,
	IserverMsgSubType.FIXEDINCOMEMARKETDATAINCREMENTAL.value: FixedIncomeMarketDataIncrementalRefresh,
	IserverMsgSubType.ALLOCINFO_RESPONSE.value: AllocationInfoResponse,
	IserverMsgSubType.EXECUTION_REQUEST.value: ExecutionRequest,
	IserverMsgSubType.CLIENT_SETTINGS_UPDATE.value: ClientSettingsUpdate,
	IserverMsgSubType.DELETE_GROUP.value: DeleteGroup,
	IserverMsgSubType.OMSSETTINGS.value: OMSSettings,
	IserverMsgSubType.CHANGEPASSWORD.value: ChangePasswordRequest,
	IserverMsgSubType.GET_ORDERS_REQUEST.value: GetOrdersRequest,
	IserverMsgSubType.BROADCAST.value: Broadcast,
	IserverMsgSubType.LOGOUT.value: Logout,
	IserverMsgSubType.DELETE_ACCOUNT.value: DeleteAccount,
	IserverMsgSubType.PAIRS_SPREAD_ORDER_RESPONSE.value: PairsSpreadOrderResponse,
	IserverMsgSubType.QUOTE_REQUEST.value: QuoteRequest,
	IserverMsgSubType.TEST_CHILD_CLASS.value: TestChildClass,
	IserverMsgSubType.IOI.value: IOI,
	IserverMsgSubType.CLIENT_ORDER_REQUEST.value: ClientOrderRequest,
	IserverMsgSubType.LOGON.value: LogonRequest,
	IserverMsgSubType.TEST_PARENT_CLASS.value: TestParentClass,
	IserverMsgSubType.PARENT_ORDER_RESPONSE.value: ParentOrderResponse,
	IserverMsgSubType.ORDERRESPONSE.value: OrderResponse,
	IserverMsgSubType.FORGOTPASSWORD_RESPONSE.value: ForgotPasswordResponse,
	IserverMsgSubType.QUOTE.value: Quote,
	IserverMsgSubType.MOVE_ACCOUNT.value: MoveAccount,
	IserverMsgSubType.COMPANY_SETTINGS_UPDATE.value: CompanySettingsUpdate,
	IserverMsgSubType.STATUS.value: EngineStatus,
	IserverMsgSubType.ALLOCATION_REQUEST.value: AllocationRequest,
	IserverMsgSubType.ADMIN_COMMAND.value: AdminCommand,
	IserverMsgSubType.CANCEL_ALL.value: CancelAll,
	IserverMsgSubType.PAIRS_SPREAD_ORDER.value: PairsSpreadOrder,
}


from .Destination import Destination
from .BondExecutionStatus import BondExecutionStatus
from .GroupAccount import GroupAccount
from .ChoiceGroup import ChoiceGroup
from .RegisteredRep import RegisteredRep
from .ChoiceItem import ChoiceItem
from .Properties import Properties
from .OrderStatus import OrderStatus
from .AllocationRequestInfo import AllocationRequestInfo
from .ExecutionStatus import ExecutionStatus
from .OrderRequestInfo import OrderRequestInfo
from .AllocAccountInfo import AllocAccountInfo
from .ClientOrderRequestInfo import ClientOrderRequestInfo
from .ParentOrderStatus import ParentOrderStatus
from .AllocationRequestStatus import AllocationRequestStatus
from .AllocInfo import AllocInfo
from .Settings import Settings
from .PartiesStatus import PartiesStatus
from .AccountInfo import AccountInfo
from .YieldData import YieldData
from .GroupAccountChange import GroupAccountChange
from .AllocationStatus import AllocationStatus
from .OrderRequestStatus import OrderRequestStatus
from .IncrementalMdEntries import IncrementalMdEntries
from .ClientOrderStatus import ClientOrderStatus
from .StrategyInfo import StrategyInfo
from .UpDown import UpDown
from .QuoteRecord import QuoteRecord
from .OfficeGroupClient import OfficeGroupClient
from .Strategy import Strategy
from .RiskProfile import RiskProfile
from .TraderList import TraderList
from .SnapshotMdEntries import SnapshotMdEntries
from .TradeUser import TradeUser


msgs_by_block_tag = {
	'DESTSTRATBL': Destination,
	'BEB': BondExecutionStatus,
	'GRPACC': GroupAccount,
	'CHOGBLK': ChoiceGroup,
	'REGREP': RegisteredRep,
	'CHOIBLK': ChoiceItem,
	'PRPSBLK': Properties,
	'OB': OrderStatus,
	'ALLOCREQBLOCK': AllocationRequestInfo,
	'EB': ExecutionStatus,
	'ORB': OrderRequestInfo,
	'ALLOCABLK': AllocAccountInfo,
	'CORDREQBLOCK': ClientOrderRequestInfo,
	'PB': ParentOrderStatus,
	'ALLOCREQSB': AllocationRequestStatus,
	'ALLOCINFO': AllocInfo,
	'SETTBLK': Settings,
	'PRTS': PartiesStatus,
	'ACCTINFO': AccountInfo,
	'YDB': YieldData,
	'GRPACCEDT': GroupAccountChange,
	'ALLOCSTATBLOCK': AllocationStatus,
	'ORSB': OrderRequestStatus,
	'INCREMENTALMDENTRIESBLOCK': IncrementalMdEntries,
	'CORDSTATBLOCK': ClientOrderStatus,
	'STGBLK': StrategyInfo,
	'UPDWNBLK': UpDown,
	'QUOTE': QuoteRecord,
	'OGCBLK': OfficeGroupClient,
	'STGBLK': Strategy,
	'RISKPROFILE': RiskProfile,
	'TRLST': TraderList,
	'SNAPSHOTMDENTRIESBLOCK': SnapshotMdEntries,
	'TRDUSER': TradeUser,
	'STRATTVS': TagValueMsg,
	'STRATDEFTVS': TagValueMsg,
	'STRATSETTVS': TagValueMsg,
}


block_tag_collisions = {
	'STGBLK' : {
		ClientOrderRequest: StrategyInfo,
		ClientOrderResponse: StrategyInfo,
		Destination: Strategy,
		OrderRequestInfo: StrategyInfo,
		OrderResponse: StrategyInfo,
	}

}



'''
 Generated by EZX Code Generator  
Date: 04/18/22 
'''


from iserver.EzxMsg import EzxMsg, EzxMsgMeta
from iserver.msgs import *
from iserver.enums import api
from iserver.msgs.ClientOrderRequestInfo import ClientOrderRequestInfo

from .StrategyInfo import StrategyInfo



class ClientOrderRequest(ClientOrderRequestInfo, metaclass=EzxMsgMeta):

	__block_fields__ = {
		'STGBLK' : Field('strategyInfo', 'STGBLK', noop_decode_func, ), 
	}

	def __init__(self, **kwargs):
		self.strategyInfo : StrategyInfo = None
		super(ClientOrderRequest, self).__init__(**kwargs)
		self.msg_type = api.IserverMsgType.APPLICATION.value
		self.msg_subtype = api.IserverMsgSubType.CLIENT_ORDER_REQUEST.value




'''
Contains the API field descriptors and mappings to API tags and data types
Created on Sep 29, 2021

@author: Sgoldberg
'''
from iserver.EzxMsg import EzxMsg
from iserver.msgs import Field
from iserver.msgs.field_definitions import api_fields


# this is where one API tag maps to 2 different field objects
_api_tag_collisions = {}

_api_tags = {}
for field in api_fields.values():
    
    if (field.api_tag in _api_tags):  # this is an API Tag collision!
        l = _api_tag_collisions.get(field.api_tag)
        if l == None:
            l = []
            l.append(_api_tags.get(field.api_tag))
            _api_tag_collisions[field.api_tag] = l
        l.append(field)        
       
    _api_tags[field.api_tag] = field

_custom_fields_by_name = {
        'tag_values' : Field('tag_values')
    
    }    


def get_api_field(name):    
    f = api_fields.get(name)
    if not f:
        f = _custom_fields_by_name.get(name)
        
    return f



def get_msg_field(api_tag : str, msg_object : EzxMsg=None):
    try:
        # check collisions first
        fields = _api_tag_collisions.get(api_tag)
        if fields:
            # find the field object which matches a field name on the message
            return next((f for f in fields if msg_object and hasattr(msg_object, f.name)), None)
        # normal case
        return _api_tags[api_tag]
    except:
        return 
    




import random
import string
from test.pickletester import metaclass

from iserver.msgs import *
from iserver.msgs import msgtags


class MyMetaClass(type):
    def __new__(cls, clsname, bases, attrs):
        print(f'\ncls={cls}, clsname={clsname}, attrs={attrs}')
        if '__block_fields__' in attrs:
            attrs['__block_fields_by_name__'] = {}
            for field in attrs['__block_fields__'].values():
                attrs['__block_fields_by_name__'][field.name] = field
                print(f'added field={field}')
            
        
        return type(clsname, bases, attrs)
    


class MsgWithBlockFields(object, metaclass=MyMetaClass):
    
        #make static/class level collection since this does not vary by instance.
    __block_fields__ = {
        'EB' : Field('executions', 'EB', noop_decode_func , update_list_item)
        }
    

class Parent(object, metaclass=MyMetaClass):
    pass

class Child(Parent):
    def __init__(self):
        print('Child().__init__()')
        super(Child, self).__init__()

class GrandChild(Child, metaclass = MyMetaClass):
    
        #make static/class level collection since this does not vary by instance.
    __block_fields__ = {
        'EB' : Field('executions', 'EB', noop_decode_func , update_list_item)
        }
    
    def __init__(self):
        print('GrandChild().__init__()')        
        super(GrandChild, self).__init__()
        

def random_symbol(max_chars : int =4) -> string:
    return ''.join(random.choices(string.ascii_uppercase, k=max_chars))

def random_price(min_price : float = .01, max_price : float = 999.99) -> float:
    return round(random.uniform(min_price, max_price), 4)

def random_quantity(min_qty : int = 1, max_qty : int = 5000):
    return random.randint(min_qty, max_qty)

    


from setuptools import setup, find_packages
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


setup(
    name='ezx-pyapi',
    version='0.4.2',
    license='MIT',
    author="EZX Inc.",
    author_email='support@ezxinc.com',
    packages=find_packages('.'),
    package_dir={'': '.'},
    url='http://www.ezxinc.com/',
    keywords='EZX iServer API trading',
    install_requires=[],
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires=">=3.8, <4",
    project_urls={'Sample API Application' : 'https://github.com/EZXInc/ezx-sample-py' }
)





'''
 Generated by EZX Code Generator  
Date: 04/18/22 
'''


from iserver.EzxMsg import EzxMsg, EzxMsgMeta
from iserver.msgs import *
from iserver.enums import api

from .SnapshotMdEntries import SnapshotMdEntries



class MarketDataSnapshot(EzxMsg, metaclass=EzxMsgMeta):

	__block_fields__ = {
		'SNAPSHOTMDENTRIESBLOCK' : Field('noMdEntires', 'SNAPSHOTMDENTRIESBLOCK', noop_decode_func, update_list_item), 
	}

	def __init__(self, **kwargs):
		self.fixEngineName : str = None
		self.mdMsgID : str = None
		self.symbol : str = None
		self.engineSeqno : int = None
		self.sendingTime : str = None
		self.msgRecvTime : str = None
		self.mdReqID : str = None
		self.text : str = None
		self.noMdEntires : list[SnapshotMdEntries] = None
		super(MarketDataSnapshot, self).__init__(**kwargs)
		self.msg_type = api.IserverMsgType.APPLICATION.value
		self.msg_subtype = api.IserverMsgSubType.MARKETDATASNAPSHOT.value




'''
Created on Sep 28, 2021

@author: Sgoldberg
'''
import unittest
import io

class Test(unittest.TestCase):


    def setUp(self):
        pass


    def tearDown(self):
        pass


    def testThisWorks(self):
        self.assertTrue(True, "should be true")

    def testStringBuffer(self):
        buffer = io.StringIO()
        buffer.write("Hello ")
        buffer.write("World")
        s = buffer.getvalue()
        self.assertEqual("Hello World", s, "got it")
        

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testThisWorks']
    unittest.main()


'''
 Generated by EZX Code Generator  
Date: 04/18/22 
'''


from iserver.EzxMsg import EzxMsg, EzxMsgMeta
from iserver.msgs import *
from iserver.enums import api




class IncrementalMdEntries(EzxMsg, metaclass=EzxMsgMeta):


	def __init__(self, **kwargs):
		self.mdentryRefID : str = None
		self.symbol : str = None
		self.mdentryType : str = None
		self.mdentryPrice : float = None
		self.msgRecvTime : str = None
		self.mdentrySize : float = None
		self.mdentryPositionNumber : int = None
		self.mdentryOriginator : str = None
		self.quoteCondition : str = None
		self.quoteEntryID : str = None
		self.ezxMdId : str = None
		self.mdUpdateAction : str = None
		super(IncrementalMdEntries, self).__init__(**kwargs)




'''
 Generated by EZX Code Generator  
Date: 04/18/22 
'''


from iserver.EzxMsg import EzxMsg, EzxMsgMeta
from iserver.msgs import *
from iserver.enums import api




class UpDown(EzxMsg, metaclass=EzxMsgMeta):


	def __init__(self, **kwargs):
		self.updownName : str = None
		self.decimalPlaces : int = None
		self.initialValue : float = None
		self.increment : float = None
		super(UpDown, self).__init__(**kwargs)




'''
Created on Oct 20, 2021

@author: Sgoldberg
'''
from enum import Enum, auto
import logging
import socket

from iserver import ezx_msg, API_HEADER_SIZE
from iserver.EzxMsg import EzxMsg
from iserver.enums.msgenums import LogonType
from iserver.msgs import LogonRequest
from iserver.msgs import LogonResponse
import iserver
import threading
import time
from iserver.util import StopWatch
from iserver.msgs.HeartBeatMsg import HeartBeatMsg

logger = logging.getLogger(__name__)

def connect(host, port) -> socket.socket :
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect((host, port))
    logger.info(f'connected to {host}:{port}')
    return s
    
    
def empty_msg_handler(msg_subtype: int, msg: EzxMsg):
    logger.info(f"received msg={msg}")


class NotLoggedInException(Exception):
    def __init__(self, msg : str = 'Not logged into iServer'):
        super().__init__(msg)
    


class ClientState(Enum):
    INITIAL = auto()
    CONNECTING = auto()
    CONNECTED = auto()
    LOGGED_IN = auto()
    LOGON_FAILURE = auto()
    DISCONNECTED = auto()
    STOPPED = auto()



DEFAULT_CONNECT_RETRY_SECONDS = 5
DEFAULT_HEARTBEAT_SECONDS = 15 

class ConnectionInfo(object):
    '''
    Specifies all settings related to connecting to the iServer.
    '''

    def __init__(self, host : str, port : int, company : str, user : str, password :str, 
                 logon_type : int = LogonType.ALL_FILLS_MSGS_AND_ALL_OPEN_ORDERS.value, 
                 connect_retry_seconds : float = DEFAULT_CONNECT_RETRY_SECONDS, heartbeat_seconds : int = DEFAULT_HEARTBEAT_SECONDS):
        '''
        Constructor
        '''        
        self.host = host
        self.port = port
        self.company = company
        self.user = user
        self.password = password
        self.logon_type = logon_type
        self.connect_retry_seconds = connect_retry_seconds
        heartbeat_seconds = min(DEFAULT_HEARTBEAT_SECONDS, heartbeat_seconds)
        self.heartbeat_seconds = heartbeat_seconds        
        
    def __str__(self):
        return iserver.to_bean_string(self)

class ApiClient(object):
    '''
    Network client used to connect, send and receive messages to and from the iServer.
    '''
    def __init__(self, connection_info : ConnectionInfo, msg_handler = empty_msg_handler, state_handler = lambda *args: None, last_seqno : int = 0):
        '''
        Constructor
        '''        
        self.connection_info = connection_info
        self._msg_handler = msg_handler
        self._state = ClientState.INITIAL
        self.on_state_change = state_handler
        self._last_seqno = last_seqno
        self.connection_monitor = ClientMonitor(self)

        
    def start(self):        
        '''
        Start the connection with iServer
        '''        
        if self.is_running():
            return
        
        logger.info("start(): starting client")
        self._set_state(ClientState.INITIAL)
        self.connection_monitor.start()
        

    def _connect(self):
        logger.info("start(): connecting to iServer...")
        try:
            self._socket = connect(self.connection_info.host, self.connection_info.port)
            # if we get here, we're connected
            self._set_state(ClientState.CONNECTED)
            
            logger.info("start(): connected")
            self.__recv_thread = threading.Thread(target=self._recv)
            self.__recv_thread.start()
            
            self._sendLogin()
            
        except Exception as e:
            logger.error('start(): failed to connect! ex=%s', e)            
            #todo: save exception?
            
        
    def stop(self):
        '''
        Disconnect and shut down the ApiClient
        '''
        # set STOPPED state first to prevent reconnect logic
        self._set_state(ClientState.STOPPED)        
        try:
            if self._socket:
                self._socket.close()
                self._socket = None
                
                logger.info("stop(): socket closed.")
            
            if self.connection_monitor:
                self.connection_monitor.stop()
                
               
        except:
            pass
        
    def _heartbeat(self):
        self.send_message(HeartBeatMsg())
    
    def state(self) -> ClientState:
        return self._state
    
    def _set_state(self, state : ClientState):
        logger.debug("ApiClient.state=%s", state)        
        self._state = state
        self.on_state_change(state)    
        
    @property
    def last_seqno(self):
        return self._last_seqno

    @last_seqno.setter
    def last_seqno(self, seqno):
        if not seqno or seqno < self._last_seqno:
            #don't allow setting bad values
            return 

        self._last_seqno = seqno
            
    
    def is_running(self):
        '''
        Indicates that client has been started and is running        
        '''
        return self.connection_monitor.is_running() # if the monitor is running, the client is alive
    
        
    def is_stopped(self):
        '''
        Returns True when Client was stopped programmatically to shut it down.
        '''
        return self._state == ClientState.STOPPED
    
    def is_connected(self):
        return self._state == ClientState.LOGGED_IN or self._state == ClientState.CONNECTED    
    
    def is_loggedin(self):
        return ClientState.LOGGED_IN == self.state()
    
    def _sendLogin(self):
        logger.info("sendLogin(): sending logon message now.")
        self.send_message(_create_login(self.connection_info, self.last_seqno))
        
    def _handle_message(self, msg_type : int, msg_subtype : int, body : bytearray):
        body = body.decode('UTF-8')
        logger.debug(f"_handle_message(): processing subtype={msg_subtype}, msg={body}")
        msg_object = ezx_msg.decode_message(msg_subtype, body)
                
        if msg_object:
            # handle the logon response first.
            if type(msg_object) is LogonResponse:  
                self.handle_logon_response(msg_object)
            
            try:
                # pass message to the user            
                self._msg_handler(msg_subtype, msg_object)
                self.last_seqno = msg_object.get_seqNo()
                
                
            except Exception as e:  #catching Exception here prevents an error in the user code from killing the receive thread.
                logger.error('exception processing msg=%s. ex=%s', msg_object, e)
                
        else:                        
            logger.warn(f"_handle_message(): did not decode subtype={msg_subtype}")         
        
    def handle_logon_response(self, logon_response):
        if logon_response.returnCode == 0:
            self._set_state(ClientState.LOGGED_IN)
        else:
            logger.fatal(f"handle_logon(): logon failed! returnCode={logon_response.returnCode}, failure message={logon_response.returnDesc} ")            
            logger.fatal('handle_logon(): stopping client. Please contact EZX support!')
            self.logon_failure = (logon_response.returnCode, logon_response.returnDesc)
            self._set_state(ClientState.LOGON_FAILURE)
   
            self.stop()
        
    def _recv(self):
        if not self._socket:
            logger.error("_recv(): no socket to receive, exiting")
            return 
        
        # s = self._socket
        with self._socket:
            header = bytearray(API_HEADER_SIZE) 
            header_recv_length = 0
            body_recv_length = 0
            while self._socket:
                try:
                    logger.debug('_recv(): waiting for message...')
                    if header_recv_length < API_HEADER_SIZE:
                        # use memoryview to create a view of the header buffer from last byte read until the end.
                        header_recv_length += self._socket.recv_into(memoryview(header)[header_recv_length:])
                        
                        if header_recv_length <= 0: 
                            # recv_into seems to have a glitch. It normally blocks while waiting for data. However, 
                            # if client gets disconnected from the server, it does not seem to raise a ConnectionAborted error,
                            # and instead returns with 0 bytes read immediately and causes this method to loop indefinitely.
                            logger.warn("_recv(): recv_into return 0 bytes - indicates disconnect?")
                            raise ConnectionAbortedError                                        
                        
                    if header_recv_length < API_HEADER_SIZE:
                        logger.debug('_recv(): received partial header. expected=%d, received=%d', API_HEADER_SIZE, header_recv_length)
                        continue    # wait to receive the rest of the header
                    
                    length, msg_type, msg_subtype = ezx_msg.parse_header(header)
                    
                    #TO-DO: wrap this in if statement
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug('_recv(): got header. reading body of length=%d. (header=%s)', length, header.decode('UTF-8'))
                        
                    if body_recv_length == 0:
                        body = bytearray(length)

                    # use memoryview to create a view of the body buffer from last byte read until the end.                         
                    body_recv_length += self._socket.recv_into(memoryview(body)[body_recv_length: ])
                    if body_recv_length < length:
                        logger.debug('_recv(): received partial message. expected=%d, received=%d - waiting for remainder of the message', length, body_recv_length)
                        continue
                    
                    header_recv_length = 0
                    body_recv_length = 0
                    
                    #TO-DO: wrap this in if statement
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug('_recv(): got msg, processing msg type=%d. msg=%s', msg_type, body.decode('UTF-8'))
                    
                    self._handle_message(msg_type, msg_subtype, body)
                    
                except ConnectionAbortedError:
                    #TODO: figure out whether this was user disconnect or from another source.
                    logger.info("_recv(): socket was disconnected")
                    break
                    
                except Exception as e:
                    logger.exception(f"_recv(): error receiving message! ex={e}")
                    break;

        logger.debug("_recv(): no longer receiving messages.")

        self._socket = None

        if not self.is_stopped():
            self._set_state(ClientState.DISCONNECTED)
         
    def send_message(self, api_msg_object: EzxMsg):
        
        # TODO: do we need thread synchronization around the socket.sendall method?  Also maybe synchronize the state() and _set_state() methods
        # to this same lock?        
        
        if not self.is_loggedin() and not type(api_msg_object) == LogonRequest: # LogonRequest is sent before being in the logged in state
            raise NotLoggedInException()
                        
        try:
            msg = ezx_msg.to_api_bytes(api_msg_object)
            self._socket.sendall(msg)
            logger.debug("send_message(): sent msg=%s", msg)
            
        except Exception as e:
            logger.error('send_message(): error on send - probably a disconnect! ex=%s', e)
    
    
         
    
    def get_logon_failure(self) -> (int, str):
        '''
        Retrieve logon failure code and description if iServer indicates that the Logon failed
        @return: tuple of errcode (int) and message 
        '''
        try:
            return self.logon_failure
        
        except AttributeError: # this might not be set
            pass
            
       
        
class ClientMonitor(object):
    
    
    def __init__(self, client, wait_seconds = DEFAULT_CONNECT_RETRY_SECONDS, heartbeat_seconds = DEFAULT_HEARTBEAT_SECONDS):
        self.client = client
        self.monitor_thread = None
        self.heartbeat_timer = StopWatch()

    def retry_seconds(self) -> float: 
        return self.client.connection_info.connect_retry_seconds
    
    def heartbeat_seconds(self) -> float:
        return self.client.connection_info.heartbeat_seconds    
    
    def start(self):
        if self.monitor_thread:
            logger.warn('started called when ClientMonitor is already running.')
            return 

        logger.info("ClientMonitor.start(): starting monitor thread")
        self.run = True        
        self.monitor_thread = threading.Thread(target=self.monitor)
        self.monitor_thread.daemon = True                        
        self.monitor_thread.start()
    
        
    def stop(self):
        logger.info("ClientMonitor.stop(): shutting down monitor")
        self.run = False
                        
        
    def monitor(self):
        logger.info("ClientMonitor.monitor(): starting connection monitor")        
        # 1. if not connected, connect
        # 2. if logged in, send heartbeat
        while self.run:
            if self.connect_required():
                self.client._connect()
                time.sleep(self.retry_seconds())
                self.heartbeat_timer.start()            
                
            elif ClientState.LOGGED_IN == self.client.state() and self.heartbeat_timer.elapsed_seconds() >= self.heartbeat_seconds():
                self.client._heartbeat()
                self.heartbeat_timer.start()                
                                
        logger.info('ClientMonitor.monitor(): exiting')
        self.monitor_thread = None
    
    def connect_required(self):
        return self.client.state() == ClientState.INITIAL or self.client.state() == ClientState.DISCONNECTED
    
    def is_running(self):
        if self.monitor_thread:
            return self.monitor_thread.is_alive()



def _create_login(info : ConnectionInfo, seqno : int = 0):
    l = LogonRequest()
    l.companyName = info.company
    l.userName = info.user
    l.password = info.password
    l.logonType = info.logon_type
    # to-do handle setting seqno
    l.seqNo = seqno
    return l


'''
Created on Oct 6, 2021

@author: Sgoldberg
'''
import iserver


'''
Metaclass for all EzxMsg subclasses to implement whatever common setup code is needed
'''
class EzxMsgMeta(type):
    
    '''
        optionally create the __block_fields_by_name dictionary if __block_fields__ is defined
    '''
    def __new__(cls, clsname, bases, attrs):        
        if '__block_fields__' in attrs:
            attrs['__block_fields_by_name__'] = {}
            for field in attrs['__block_fields__'].values():
                attrs['__block_fields_by_name__'][field.name] = field                
            
        
        return type(clsname, bases, attrs)


'''
classdocs
'''
class EzxMsg(object):

    def __init__(self, **kwargs):
        iserver.set_properties(self, kwargs)
    
    def __eq__(self, other):
        return iserver.msg_equals(self, other)
        
    def __str__(self):
        return iserver.to_bean_string(self)
    
    def __repr__(self):
        return iserver.to_bean_string(self, iserver._all_props_filter)
    

    def get_block_field(self, api_block_tag : str):
        '''
        When decoding a received message, if there is a "block" found in the message, get a Field for the specified block tag. Returns none
        if this EzxMsg subtype does not contain a field for this block.
        
        @param api_block_tag: the API tag name which corresponds to the block field we are looking for
        @return: the Field definition or None if not found
        '''
                
        found = self.do_block_field_lookup(api_block_tag)
        if found:
            return found[1]
                

    def get_block_field_by_name(self, field_name : str):
        '''
        When encoding, we need to see whether the specified messge field (field_name parameter)
        is a block field -- in which it is encoded as a message, or whether it is a simple field --
        in which case it is encoded as single text value.  Will return None if the field_name is 
        not a block field    
        '''        
        # so ... the field we are looking for could be defined in a super class of the current instance
        # therefore we iterate the class heirarchy of the object to check each class for existence of the field we are looking
        # for.        
        mro = self.__class__.__mro__
        for t in mro:            
            try:
                f = t.__block_fields_by_name__.get(field_name)
                if f:
                    return f
            except AttributeError:
                pass
        
    
    def get_block_field_owner(self, api_block_tag: str):
        found = self.do_block_field_lookup(api_block_tag)
        if found:
            return found[0]
 
            

    def do_block_field_lookup(self, api_block_tag: str):
        # so ... the field we are looking for could be defined in a super class of the current instance
        # therefore we iterate the class heirarchy of the object to check each class for existence of the field we are looking
        # for.        
        mro = self.__class__.__mro__
        for t in mro:            
            try:
                f = t.__block_fields__.get(api_block_tag)
                if f:
                    return t,f
            except AttributeError:
                pass
            
    
    def update(self, tag : str, value : object):
        '''
        Optional method to update this message object from the tag/value pair. This would be used for custom messages
        which do not have their fields identified in the iserver.msgs.field_definitions module.
        
        @param tag: API tag which identifies the field to update
        @param value: value to assign t the field  
        '''
        pass # no-op, mean to be overridden in subclasses
    
    def get_seqNo(self) -> int:
        '''
        Gets the API sequence number of this message. 
        '''
        if hasattr(self, 'seqNo'):  #seqNo is not defined in all API messages, but when it is, it should be stored
            return self.seqNo
        
    


'''
 Generated by EZX Code Generator  
Date: 04/18/22 
'''


from iserver.EzxMsg import EzxMsg, EzxMsgMeta
from iserver.msgs import *
from iserver.enums import api




class OrderRequestStatus(EzxMsg, metaclass=EzxMsgMeta):


	def __init__(self, **kwargs):
		self.version : int = None
		self.reqState : str = None
		self.myID : str = None
		self.basketID : str = None
		self.clientOrderID : str = None
		self.importFileName : str = None
		self.importTimeStamp : str = None
		self.importJobName : str = None
		self.ezxClientOrderID : int = None
		self.ezxClientOrderVersion : int = None
		super(OrderRequestStatus, self).__init__(**kwargs)




import logging

logger = logging.getLogger(__name__)


class Field:
    '''
    Field descriptor for API message fields.  API messages are sent as strings of name/value pairs. The name is referred to as the API tag.
    
    Attributes:
        api_tag (str): The API tag used for this field. For example:  DEST (for the destination field)
        from_string (function): a function which can convert the string representation of the value to the correct data type. Used for decoding received messages. Default= str
        update_func (function): a function to update the corresponding attribute in the message with the field value. Default = setattr. The other function used is msgs.update_list_item
            if the message attribute is a list object.
    '''

    
    def __init__(self, name=None, api_tag=None, decode_func=str, update_func=setattr):
        '''
        Constructor 
        
        Parameters:
            name (str): Field name/attribute name in the message class
            api_tag (str): the API tag string which identifies the field in the message
            
        
        '''
        self.name = name
        self.api_tag = api_tag
        self.from_string = decode_func
        self.update_func = update_func
                
    def __str__(self):
        return f'Field: name={self.name}, api_tag={self.api_tag}'
    
    def decode_value(self, value):
        return self.from_string(value)
    
    def update_msg_field(self, msg_obj, value):
        # setattr(m, msg_field.name, value)
        self.update_func(msg_obj, self.name, self.decode_value(value))


def noop_decode_func(value):
    '''
        Placeholder function for fields that don't have a decode logic
    '''    
    return value


def update_list_item(obj, property_name, item_value):
    '''
    Fields which are lists of objects use this as their update function.  This will add the specified item_value to the list specified by property_name.
    '''        
    l = getattr(obj, property_name)
    if not l:
        l = []
        setattr(obj, property_name, l)    
    l.append(item_value)    


__all__ = [ 'Field','noop_decode_func' ,'update_list_item' ]

# do not move these import statements
from .ClientOrderResponse import ClientOrderResponse
from .ClientOrderStatus import ClientOrderStatus
from .ExecutionStatus import ExecutionStatus
from .LogonRequest import LogonRequest
from .LogonResponse import LogonResponse
from .OrderResponse import OrderResponse
from .OrderStatus import OrderStatus



